/*****************************************************************************
 * $LastChangedDate: 2008-11-29 14:20:16 -0500 (Sat, 29 Nov 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Module configuration.
 *//*
 * LEGAL:	COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define BASE_CONF_CC 1
#include "base/module.hh"
#include "base/conf.hh"
#include "base/file.hh"

namespace base {

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  Conf  ///////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

INTERN char  sDummyArg0[1] = { 0 };
INTERN char* sDummyArgv[] = { sDummyArg0 };

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Conf::Conf( void )
:	mValid(false),
	mProgramName(""),
	mArgc(0),  // mArg* are assigned at Init()
	mArgv(sDummyArgv),
	mArgv0(""),  // C++ string
	// cmd-line args
	mArgVerbose(false),
	mArgQuiet(false),
	// --- .conf file ---
	mVerbose(true),
	mQuiet(false),
	mRandomSeed(0x2b3e27bd)
{
}

/*****************************************************************************
 * Initialize/read configuration.
 *****************************************************************************/
void
Conf::Init( int argc, char** argv, const string& programName, const string& confTxtFilename )
{

BEGIN_ONCE {

	mValid = true;  // assign this first as *_CONF may expand to GetConf() which asserts it
    BASE_CONF.mProgramName = programName;  // first of all

	ParseCmdlineArgs( argc, argv );   // cmdline args parsed prior to reading .conf file
	ReadConfTxt( confTxtFilename );   // read from textual .conf file
	ParseCmdlineArgs2( argc, argv );  // cmdline args that override .conf file

} END_ONCE

}

/*****************************************************************************
 * Parse cmdline args.
 *****************************************************************************/
void
Conf::ParseCmdlineArgs( int argc, char** argv )
{
    // Subtle: -v/-q cmdline overrides .conf.
    // ConfTxt::ProcessFile() ignores Verbose/Quiet conf lines if those args were passed.

	// Some args need to be parsed very early, before .conf is read.
	for ( int i = 1; i < argc; ++i )
	{
		if ( STREQ( argv[i], "-q" ) )  // override .conf
		{
			mArgQuiet   = mQuiet   = true;
			mArgVerbose = mVerbose = false;
			CmdargDone( i );
		}
		else if ( STREQ( argv[i], "-v" ) )  // override .conf
		{
			mArgQuiet   = mQuiet   = false;
			mArgVerbose = mVerbose = true;
			CmdargDone( i );
		}
#if DEBUG
		else if ( STREQ( argv[i], "-core" ) )  // dump core instead of throwing exception
		{
			gGlobals.mAssertAbort = true;
			CmdargDone( i );
		}
#endif
	}
}

void
Conf::ParseCmdlineArgs2( int argc, char** argv )
{
	// NOP
}

/*****************************************************************************
 * Read from .conf file.
 *****************************************************************************/
void
Conf::ReadConfTxt( const string& confTxtFilename )
{
	try
	{
		ModuleConfTxt confTxt;
		confTxt.Read( string(base::GetDataDir()) + string(confTxtFilename) );
	}
	catch (...)
	{
		CERROR << "ERROR: cannot read " << confTxtFilename << endl;
		EXIT( 1 );
	}
}

void
Conf::ModuleConfTxt::ProcessField( string& field, ifstream& conf )
{
	bool error = false;
	if ( field == "Verbose" )  // -v/-q override .conf
	{
		if ( BASE_CONF.mArgQuiet or BASE_CONF.mArgVerbose )
		{
			int ignore; conf >> ignore;
		}
		else
		{
			conf >> BASE_CONF.mVerbose;
		    if ( BASE_CONF.mVerbose and BASE_CONF.mQuiet )
				error = true;
		}
	}
	else if ( field == "Quiet" )  // -v/-q override .conf
	{
		if ( BASE_CONF.mArgQuiet or BASE_CONF.mArgVerbose )
		{
			int ignore; conf >> ignore;
		}
		else
		{
			conf >> BASE_CONF.mQuiet;
		    if ( BASE_CONF.mVerbose and BASE_CONF.mQuiet )
				error = true;
		}
	}
	else if ( field == "RandomSeed" )
	{
		conf.setf( ios_base::hex, ios_base::basefield );
		conf >> BASE_CONF.mRandomSeed;
		conf.setf( ios_base::dec, ios_base::basefield );
	}

    // Errors?
	if ( error )
	{
		CERROR << "ERROR[conf]: parameter for " << field << " is invalid." << endl;
		EXIT( 1 );
	}
}

/*****************************************************************************
 * Print Conf configuration.
 *****************************************************************************/
ostream&
operator<<( ostream& strm, const Conf& conf )
{
	std::ios::fmtflags savedFlags = strm.flags();

	strm << "conf [base]:"                                                   << endl
         << "RandomSeed                  = 0x" << hex << conf.mRandomSeed; //<< endl;

	strm.flags(savedFlags);

	return strm;
}

////////////////////////////////////////////////////////////////////////////////
///////////////////////////////  functions  ////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * @return Directory pathname containing the data/ subdirectory.
 * @pre BASE_CONF.mProgramName was assigned.
 *****************************************************************************/
const string&
GetDataDir( void )
{
	const string& programName = BASE_CONF.mProgramName;

	// programName is required to determine pathname of data dir.
	if ( programName.empty() )
	{
		CERROR << "BASE_CONF.mProgramName UNDEFINED!" << endl;
		ASSERT(false);
	}

    // Find directory once.
    PERSISTENT string sDataDir;
    if ( sDataDir.empty() )
    {
    	sDataDir = "data/";
    	if ( IfDirExists( sDataDir ) )
            return sDataDir;

		// If run at CWD=ProgramName/modules/sim (unusual).
    	sDataDir = "../../data/";
    	if ( IfDirExists( sDataDir ) )
            return sDataDir;

		// If run at CWD=ProgramName/modules/sim/out (unusual).
    	sDataDir = "../../../data/";
    	if ( IfDirExists( sDataDir ) )
            return sDataDir;

		// Linux:
		sDataDir = string("/usr/local/games/") + string(programName) + string("/data/");
    	if ( IfDirExists( sDataDir ) )
            return sDataDir;

		sDataDir = string("/usr/share/games/") + string(programName) + string("/data/");
    	if ( IfDirExists( sDataDir ) )
            return sDataDir;

		sDataDir = string("/usr/games/") + string(programName) + string("/data/");
    	if ( IfDirExists( sDataDir ) )
            return sDataDir;

		// FreeBSD:
		sDataDir = string("/usr/local/share/games/") + string(programName) + string("/data/");
    	if ( IfDirExists( sDataDir ) )
            return sDataDir;

		// Mac OS X fink:
		sDataDir = string("/sw/share/") + string(programName) + string("/data/");
    	if ( IfDirExists( sDataDir ) )
            return sDataDir;

        CERROR << "ERROR: Couldn't find data/ directory." << endl;
        EXIT( 1 );
    }

    return sDataDir;
}

} // namespace base
