/******************************************************************************
 * $LastChangedDate: 2008-05-04 16:06:39 -0400 (Sun, 04 May 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Random number functions.
 * @verbatim
 *
 * Random number methods are per-thread.
 * ANSI C rand_r(seed) is passed a thread variable.
 *
 * Avoid mixing BSD random() functions.
 * POSIX random functions aren't thread-safe.
 *
 * @endverbatim
 *//*
 * LEGAL:	COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 ******************************************************************************/

#define BASE_RANDOM_CC 1
#include "base/module.hh"
#include "base/random.hh"
#include "base/conf.hh"

namespace base {

// Use #define gGlobal.
namespace global { Random randomGenerator; }

////////////////////////////////////////////////////////////////////////////////
///////////////////////////////  Random  ///////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Random::Random( void )
:	mSavedSeed(0),
	mFastTableIdx(0)
{
ASSERT_RUN_ONCE;  // Random is meant to exist only in Global
ASSERT( IfPow2( FAST_TABLE_IDX_MASK+1 ) );  // check bit mask used to wrap table index

	// Precompute random numbers for the fast functions.
	for ( uint i = 0; i < FAST_TABLE_CNT; ++i )
	{
		mFastTable[i] = random_ui();
	}
}

Random::~Random()
{
ASSERT_RUN_ONCE;
}

/*****************************************************************************
 * Initialize random number generator.
 *****************************************************************************/
void
Random::Init( void )
{
	// Randomize (set random seed).
	// If RandomSeed=0 in .conf, that means to pick a truly random seed.
	if ( BASE_CONF.mRandomSeed != 0 )
		Random::randomize( BASE_CONF.mRandomSeed );
	else
		Random::randomize( Random::random_ui_true() );
}

/*****************************************************************************
 * Seed the sequence of psuedo-random numbers.
 *****************************************************************************/
void
Random::randomize( uint seed )
{
	// rand_r() will change mSeed, so save original seed for GetSeed().
	*mSeed = uint0(seed);
	mSavedSeed = seed;
}

void
Random::randomize( void )
{
	randomize( random_ui_true() );
}

/*****************************************************************************
 * @return A psuedo-random unsigned int in range {0,..,RAND_MAX}.
 *****************************************************************************/
uint
Random::random_ui( void )
{
	uint* seed = reinterpret_cast<uint*>( mSeed.PTR() );  // PerThreadPtr<uint0*> --> uint*
#if OS_WINDOWS
    return rand();
#else
	return rand_r( seed );  // SIDE-EFFECT: will change mSeed
#endif
}

/*****************************************************************************
 * @return A psuedo-random unsigned int in range {0,..,end-1}.
 * end-1 is for picking a random-but-valid index within a container
 * by passing size of container as end.
 * random_ui(100) will return {0,..,99} (never 100).
 *****************************************************************************/
uint
Random::random_ui( uint end )
{
	// Prevent divide error exception.  Extreme compiler optimize flags skips such checks.
	if ( EX( end ) )
		return random_ui() % end;
	else
		return 0;
}

/*****************************************************************************
 * @return A psuedo-random float in range {0,..,1} or {0,..,range}.
 *****************************************************************************/
fp
Random::random_f( void )
{
	//return fp(random_ui()) / fp(RAND_MAX);	// float division is slow
	return fp(random_ui()) * fp(1.0 / RAND_MAX);
}

fp
Random::random_f( fp range )
{
	//return fp(random_ui()) / fp(RAND_MAX) * range;	// float division is slow
	return fp(random_ui()) * fp(1.0 / RAND_MAX) * range;
}

/*****************************************************************************
 * @return A true random unsigned integer.
 * @remarks This function is slow.
 * An alternative is to call randomize( random_true() ) once and thereafter
 * call the pseudo-random functions.
 *****************************************************************************/
uint
Random::random_ui_true( void )
{
#if ! OS_LINUX
	srand( uint( time(0) ) );
	return rand();
#else
	PERSISTENT int fd = -1;
	uint val32 = 0;
	if ( fd == -1 ) fd = open( "/dev/random", O_RDONLY|O_NONBLOCK );
	if ( fd == -1 || read( fd, &val32, sizeof(val32) ) != sizeof(val32) )
	{
		// Failed to open /dev/random.  Try an alternative.
		DEBUG_CODE( CDEBUG << "can't open or read /dev/random (" << strerror(errno) << ")" << endl; )
		srand( uint( time(0) ) );
		val32 = rand();
	}
	return val32;
#endif
}

} // namespace base
