/*****************************************************************************
 * $LastChangedDate: 2008-08-31 18:25:27 -0400 (Sun, 31 Aug 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Streams.
 *//*
 * LEGAL:	COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef BASE_STREAM_HH
#define BASE_STREAM_HH 1

#include <streambuf>
#include <fstream>

namespace base {

/// Abbrevations for per-thread streams.
#define COUT   (*base::gGlobals.mStreams.mOut)    // normal messages
#define CERROR (*base::gGlobals.mStreams.mError)  // error messages
#define CLOG   (*base::gGlobals.mStreams.mLog)    // unimportant info/status message
#define CDEBUG (*base::gGlobals.mStreams.mDebug)  // debug messages
#define cout WRITE_COUT_INSTEAD
#define cerr WRITE_CERR_INSTEAD

// C++ stream field formatting.
#define OMANIP_FIELD(CHARS) std::setw(CHARS) << std::right

/// Helpers for C++ streams.

//bool StreamGood( std::ios& s );  // just use stream.good() (hi-freq)

bool
StreamOpen( fstream& fs, ios::openmode bits, const string& filename );

bool
StreamReadLine( istream& strm, string& line );

bool
StreamToNextLine( istream& strm );

bool
StreamPeek( istream& strm, const string& word );

////////////////////////////////////////////////////////////////////////////////
/// @brief Streambuf class that can write to one or two files (a confluence/fork).
///
/// This is a "streambuf" class to serve as the basis of an ostream.
/// Derived from Nicolai Josuttis's STL book.
///
/// @verbatim
/// Example:
///		SafePtr<FILE> file1 = fopen(...);
///		SafePtr<FILE> file2 = fopen(...);
///     StreamFile2 streambufFile2( file1, file2 );
///     std::ostream out( &streambufFile2 );  // ostream based on streambuf
///     out << "hello" << endl;
/// @endverbatim
///
class StreambufFile2 : public streambuf
{
PREVENT_COPYING(StreambufFile2)
typedef streambuf Parent;
public:
	StreambufFile2( SafePtr<FILE> file0, bool close0,
                    SafePtr<FILE> file1 = NULL, bool close1 = false );
	~StreambufFile2();

protected:
	virtual int_type	overflow( int_type c );
	std::streamsize		xsputn( const char* buf, std::streamsize n );

	SafePtr<FILE>	mFile0, mFile1;    // stream outputs to both files
	bool			mClose0, mClose1;  // true if destructor should close the file
};

////////////////////////////////////////////////////////////////////////////////
/// @brief NopStream is a indirect way to use a disabled C++ stream.
///
/// This is a "streambuf" class to serve as the basis of an ostream.
/// Derived from Nicolai Josuttis's STL book.
///
/// @verbatim
/// Example:
///     NopStreambuf nopStreamBuf;
///     std::ostream gLog( &gNopStreambuf );
/// @endverbatim
///
class NopStreambuf : public streambuf
{
PREVENT_COPYING(NopStreambuf)
typedef streambuf Parent;
public:
	NopStreambuf( void ) { }
	~NopStreambuf() { }

protected:
	virtual int_type	overflow( int_type c ) { return ~EOF; }
	std::streamsize		xsputn( const char* buf, std::streamsize n ) { return n; }
};

////////////////////////////////////////////////////////////////////////////////
/// @brief Threaded streams.
///
/// Although Streams is thread-safe, threads that concurrently write
/// to a stream diverted to stdout/stderr will result in intermixed output.
/// The responsibility of serializing output is deferred to a client
/// which would need to lock every stream statement.
///
/// Thread::Run() opens per-thread streams.
///
class Streams : private Threadable
{
public:
						Streams( void );
						~Streams();
	CLASS_METHOD void	Init( int argc, char** argv );
	void				Open( void );
	void				Close( void );
	void				SetErrorLogFname( const string& fname );
	void				SetVerbose( const bool verbose );

public:		SafePtr<ostream>	mOut;	// delete &cout;  why my program crash?
public:		SafePtr<ostream>	mError;
public:		SafePtr<ostream>	mLog;  // unimportant info/status messages
public:		SafePtr<ostream>	mDebug;
private:	NopStreambuf		mNopStreambuf;
public:		ostream				mNopStream;
private:	string				mErrorLogFname;
private:	FILE*				mErrorLogFile;
};

} // namespace base

#endif // BASE_STREAM_HH
