/*****************************************************************************
 * $LastChangedDate: 2008-08-31 19:34:22 -0400 (Sun, 31 Aug 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Timer class that invokes registered callbacks.
 *//*
 * LEGAL:   COPYRIGHT (C) 2006 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define BASE_TIMER_CC 1
#include "base/module.hh"
#include "base/types2.hh"
#include "base/timer.hh"

namespace base {

////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////  Timer  /////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

Timer* Timer::msInstance;

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Timer::Timer( void )
:	mEnabled(true),
	mBusy(false),
	mElapsedTime(0),
	mCallbacks()
{
	ASSERT_RUN_ONCE;  // singleton
}

Timer::~Timer()
{
	ASSERT_RUN_ONCE;
}

/*****************************************************************************
 * Timer can be disabled to pause/freeze program.
 *****************************************************************************/
void
Timer::Enable( const bool enable )
{
	mEnabled = enable;
}

/*****************************************************************************
 * Register a callback.
 * NOTE: Don't re-register the same callback.
 * @param	functor
 *          Callback functor.
 * @param	millisecFreq
 *			How frequently to invoke callback.
 * @param	canDisable
 *			Default is "can disable".
 *****************************************************************************/
void
Timer::Register( shptr<Functor> functor,
                 const Milliseconds millisecFreq,
                 eDisablement canDisable )
{
ASSERT( millisecFreq >= Milliseconds(MIN_MILLISEC) );
ASSERT( millisecFreq <= Milliseconds(MAX_MILLISEC) );
ASSERT( Find(functor.PTR()) == mCallbacks.end() );  // prevent re-registering

	// Insert into STL container.
	mCallbacks.insert( make_pair<SafePtr<Functor>,Callback>( functor.PTR(),
                                                             Callback( CALLBACK_REGISTERED, functor, millisecFreq, canDisable ) ) );
}

/*****************************************************************************
 * Unregister a callback functor.
 * A callback is allowed to unregister itself (see header comments).
 * Timer's iteration thru Callbacks won't be disrupted because callbacks
 * are unregistered here by merely changing a flag (container isn't affected).
 *****************************************************************************/
void
Timer::Unregister( shptr<Functor> functor )
{
	// Find callback in container.
	Callbacks::iterator iter = Find( functor.PTR() );
	if ( iter != mCallbacks.end() )
	{
		// Immediately destroy the callback unless iterating/tick is happening.
		// The next time Tick() is called, it will purge unregistered callbacks.
		Callback& callback = iter->second;
	  //ASSERT( callback.IfRegistered() );  // allow re-unregistering (not harmful)
		CHECK_TYPESIG(&callback,TYPESIG_TIMER_CALLBACK);
		if ( mBusy )
			callback.mRegistered = CALLBACK_UNREGISTERED;  // marked for destruction
		else
			mCallbacks.erase( iter );  // destroy it now
	}
	else
	{
		ASSERT(false);  // tried to unregister a callback that was never registered
	}
}

/*****************************************************************************
 * This drives the Timer class.
 * Code integration:
 * Must be called by whatever function is periodically called.
 * @param	millisecElapsed
 *			Total milliseconds between now and when program was started.
 *****************************************************************************/
void
Timer::Tick( const Milliseconds millisecElapsed )
{
// Busy iterating.
AutoFlag autoFlag( &mBusy );

	// Remember elapsed time for GetElapsedTime().
	mElapsedTime = millisecElapsed;

	// This is the C++ exception handler that will actually be invoked by C++ exception
	// during a GLUT timer-tick.  The top-level handler in mod_sim_main.cc won't be invoked,
    // since the C++ run-time cannot unwind stack frames from C code (in GLUT).
	// But adding another "try" block allows unwinding the stack from this point.
    try
    {
		// Iterate thru callbacks.
		// Call registered callbacks if their time has come.
		// Remove unregistered callbacks.
		// Clients may unregister callbacks while this is iterating!
		// Unregister() does so without disrupting this iteration.
        for ( Callbacks::iterator iter = mCallbacks.begin();
              iter != mCallbacks.end();
              /*++iter*/ )
        {
    		// Invoke this callback if its time has come.
			// Some callbacks must be called even if Timer is disabled.
			Callback& callback = iter->second;
			CHECK_TYPESIG(&callback,TYPESIG_TIMER_CALLBACK);
			if ( EX( callback.mRegistered ) )
			{
				if ( EX( mEnabled or not callback.IfCanDisable() ) )
				{
		            callback.CallIfTimeHasCome( millisecElapsed );
				}

				++iter;  // advance iterator ("for" omits it!)
			}
			else
			{
				// This callback was unregistered.  Purge it.
				// Remove it out of container (safely using STL idiom).
				Callbacks::iterator iter2 = iter;
				++iter2;					// advance iterator before
				mCallbacks.erase( iter );	// erase() invalidates iterator
				iter = iter2;
			}
        }
    }
	catch ( const Exception& exception )
	{
		PrintException( exception, "timer-tick" );
		throw;  // rethrow
	}
}

/*****************************************************************************
 * @return Iterator to registered callback or end() if unregistered.
 *****************************************************************************/
Timer::Callbacks::iterator
Timer::Find( SafePtr<Functor> functor )
{
	// Search STL map by raw ptr to Functor.
	return mCallbacks.find( functor );
}

////////////////////////////////////////////////////////////////////////////////
//////////////////////////////  Timer::Callback  ///////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 * @param	registered
 * @param	functor
 * @param	millisecFreq
 *			How frequently to invoke callback.
 * @param	canDisable
 *			Default is "can disable".
 *****************************************************************************/
Timer::Callback::Callback( const bool registered,
                           const shptr<Functor> functor,
                           const Milliseconds millisecFreq,
                           const Timer::eDisablement canDisable )
:	mRegistered(registered),
	mFunctor(functor),
    mMillisecFreq(millisecFreq),
	mMillisecCalled(NEVER_CALLED),
	mCanDisable(canDisable==CAN_DISABLE)
{
	SET_TYPESIG(this,TYPESIG_TIMER_CALLBACK);
}

// Copy ctor defined to avoid modifying-a-copy pitfall (vestige from using list).
Timer::Callback::Callback( const Callback& src )
:	mRegistered(src.mRegistered),
	mFunctor(src.mFunctor),
    mMillisecFreq(src.mMillisecFreq),
	mMillisecCalled(src.mMillisecCalled),
	mCanDisable(src.mCanDisable)
{
	SET_TYPESIG(this,TYPESIG_TIMER_CALLBACK);
}

Timer::Callback::~Callback()
{
	INVALIDATE_TYPESIG(this,TYPESIG_TIMER_CALLBACK);
}

/*****************************************************************************
 * If at or past the time to call this callback, then call it.
 *****************************************************************************/
void
Timer::Callback::CallIfTimeHasCome( const Milliseconds millisecElapsed )
{
CHECK_TYPESIG(this,TYPESIG_TIMER_CALLBACK);
ASSERT( mRegistered );

	// Time between now and last time callback was called.
	const Milliseconds delta = millisecElapsed - mMillisecCalled;
ASSERT( delta >= Milliseconds(0) );

	// If never called or past its time.
    if ( (mMillisecCalled == NEVER_CALLED) || (delta > mMillisecFreq) )
    {
        mMillisecCalled = millisecElapsed;  // remember time it was called
        (*mFunctor)( millisecElapsed );		// invoke callback
    }
}

} // namespace base
