/*****************************************************************************
 * $LastChangedDate: 2008-10-05 01:38:29 -0400 (Sun, 05 Oct 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Basic types (used rarely).
 * @verbatim
 *
 * PTR() guidelines:
 * -----------------
 * PTR() has limited uses.
 * Acceptable uses are inside frequent functions that use PTR() in their own scope.
 * PTR() opens the pitfall of dangling pointers.
 * Never store (or pass to a function that may store) what PTR() returns.
 * Be sure that what PTR() returned isn't still used after its object was destroyed.
 *
 * @endverbatim
 *//*
 * LEGAL:	COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef BASE_TYPES2_HH
#define BASE_TYPES2_HH 1

#include "base/shared_ptr.hh"

namespace base {

//==============================================================================
//==============================================================================
//                     Virtual methods allowed in this section
//==============================================================================
//==============================================================================

////////////////////////////////////////////////////////////////////////////////
/// @brief Abstract class for Iterator objects.
///
/// In itself, an abstract Iterator class is useless, because Iterator objects
/// are returned by value which precludes polymorphism.
/// Rather, its purpose is to be a reference for writing concrete Iterators
/// that all have a compatible interface.
///
/// By using Iterator as a facade, classes such as Dlist, STL containers, Array
/// can be transparently substituted without changing clients of Iterator.
///
template<typename T>
class AbstractIterator
{
public:
                    AbstractIterator( void ) { }
    virtual         ~AbstractIterator() { }

	virtual			operator bool() = 0;
	virtual T		Next( void ) = 0;
};

//==============================================================================
//==============================================================================
//                    Virtual methods forbidden in this section
//==============================================================================
//==============================================================================

// Basic types must be fast, prevent them having virtual functions.
#define virtual VIRTUAL_METHODS_IN_BASIC_TYPES_IS_TOO_SLOW

////////////////////////////////////////////////////////////////////////////////
/// @brief Iterator object for STL vectors.
///
/// Pass CONTAINER as "vector<Class>", not just "vector".
///
template<typename T,typename CONTAINER>
class STLIterator : public base::AbstractIterator<T>
{
public:
	explicit	STLIterator( CONTAINER& cont ) : mCont(cont), mIter(cont.begin()) { }
				operator bool() { return mIter != mCont.end(); }
	T			Next( void ) { return *mIter++; }
private:
	CONTAINER&						mCont;
	typename CONTAINER::iterator	mIter;
};

////////////////////////////////////////////////////////////////////////////////
/// @brief A variable with multiple values of the same type.
/// @remarks Tested by base/tests/src/multivar.cc
///
template<typename T=int>
class Multivar
{
public:
    /*****************************************************************************
     * ctor/dtor
     *****************************************************************************/
    Multivar( void ) { }
    ~Multivar() { }

    /*****************************************************************************
     * Clear all values.
     *****************************************************************************/
    void
    Clear( void )
    {
        mValues.clear();
    }

    /*****************************************************************************
     * Remove value.
     *****************************************************************************/
    void
    Remove( T value )
    {
        mValues.remove( value );
    }

    /*****************************************************************************
     * Replace all values with this value (replacement).
     *****************************************************************************/
    void
    Set( T value )
    {
        mValues.clear();
        mValues.push_back( value );
    }

    /*****************************************************************************
     * Add an additional value (additive).
     *****************************************************************************/
    void
    Add( T value )
    {
        // Add value unless present.
        if ( ! Query(value) )
    		mValues.push_back( value );
    }

    /*****************************************************************************
     * Add or remove value.
     *****************************************************************************/
    void
    Toggle( T value )
    {
        if ( Query(value) )
    		Remove( value );
    	else
    		Add( value );
    }

    /*****************************************************************************
     * @return True if multivar has this value.
     *****************************************************************************/
    bool
    Query( T value ) const
    {
        return std::find( mValues.begin(), mValues.end(), value ) != mValues.end();
    }

    /*****************************************************************************
     * @return True if multivar is empty.
     *****************************************************************************/
    bool
    IfEmpty( void ) const
    {
        return mValues.empty();
    }

private:
    std::list<T> mValues;  // multiple values of same type
};

////////////////////////////////////////////////////////////////////////////////
/// @brief Percent
/// @remarks
/// operator+=() clamps for conveniently incrementing
/// (rather than requiring every caller to duplicate clamp code).
/// Assigning an invalid Percent is never clamped: that's always an error.
///
class Percent
{
public:
            Percent( void ) : mPercent(0) { }
            Percent( uint percent ) { Set( percent ); }  // not explicit

    // Assigning an out-of-range percent is an error.
    void    operator=( uint percent ) { Set( percent ); }
            operator uint() const { ASSERT( mPercent <= 100 ); return mPercent; }

    // These clamp.
    void    operator+=( int n ) { SetClamp( int(mPercent) + int(n) ); }
    void    operator-=( int n ) { SetClamp( int(mPercent) - int(n) ); }

private:
    // Set() is strict: it will fail if percent is out-of-range.
    void    Set( uint percent )
            {
                ASSERT( percent <= 100 );
                mPercent = percent;
            }

    // SetClamp() is lenient: it will clamp out-of-range values.
    void    SetClamp( int percent )
            {
                if ( percent < 0 )
                    mPercent = 0;
                else if ( percent > 100 )
                    mPercent = 100;
                else
                    mPercent = uint(percent);
            }
private:
    uint	mPercent;
};

////////////////////////////////////////////////////////////////////////////////
/// @brief Basic string buffer class.
/// @remarks One purpose is to wrap a C++ string for sharing it.
/// For speed/convenience, to directly manipulate the C++ string and to avoid copying,
/// mBuf is exposed.  But the methods should be used whenever possible.
///
class StringBuf
{
public:
					StringBuf( void ) { }
					~StringBuf() { }
	void			Set( const char* buf )           { mBuf.assign( buf ); }
	void			Set( const char* buf, int len )  { mBuf.assign( buf, len ); }
	void			Set( const uchar* buf, int len ) { mBuf.assign( reinterpret_cast<const char*>(buf), len ); }
	void			Erase( void )				{ mBuf.erase(); }
	bool			IfEmpty( void )	const		{ return mBuf.size() == 0; }
	int				Size( void ) const			{ return mBuf.size(); }
	const string*	GetString( void ) const		{ return &mBuf; }
	const char*		GetChars( void ) const		{ return mBuf.c_str(); }
	const uchar*	GetUchars( void ) const		{ return reinterpret_cast<const uchar*>(mBuf.c_str()); }

public:
	string		mBuf;	///< public for speed/convenience but use methods if appropriate
};

////////////////////////////////////////////////////////////////////////////////
/// @brief Automatically ensures a flag is cleared before a function returns.
/// ~AutoFlag() will be invoked regardless of an unexpected early return or thown exception.
///
class AutoFlag
{
public:
	AutoFlag( SafePtr<bool> pFlag ) : mFlag(pFlag) { *pFlag = true; }
	~AutoFlag( void ) { *mFlag = false; }
private:
	const SafePtr<bool> mFlag;  ///< points to a class member typically
};

#undef virtual

} // namespace base

#endif // BASE_TYPES2_HH
