/*****************************************************************************
 * $LastChangedDate: 2008-08-31 19:34:22 -0400 (Sun, 31 Aug 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Support for checking type signatures.
 * @verbatim
 *
 * Type signatures are used to catch pointers to objects of the wrong type
 * and to ensure the memory holding object isn't corrupted.
 * Type signatures are conditionally compiled according to #define COMPILE_TYPESIGS.
 * A set of macros are provided which expand to nothing if COMPILE_TYPESIGS=0.
 * Secondary (and additional) type signatures can be used to for checking
 * the base class type signature of derived object.
 *
 * Passing the wrong type signature by mistake will be caught by the compiler
 * as the name of the type signature member is catenated with the name of
 * the enum (this is a macro trick).
 *
 * Example:
 *
 *     // Example of declaring type signature enums.
 *     enum
 *     {
 *        TYPESIG_INVALID,
 *        TYPESIG_MYCLASS,
 *     };
 *
 *     // Example of declaring and setting a type signature in a class.
 *     // The destructor invalidates the type signature in order to
 *     // catch the error of using a destroyed object by a stale pointer.
 *     class MyClass
 *     {
 *     public:
 *         MyClass()        { SET_TYPESIG(this,TYPESIG_MYCLASS); valid = false; }
 *         MyClass( int x ) { SET_TYPESIG(this,TYPESIG_MYCLASS); valid = true; value = x; }
 *         ~MyClass()       { INVALIDATE_TYPESIG(this,TYPESIG_MYCLASS); }
 *         bool valid;
 *         int  value;
 *         DECL_TYPESIG(TYPESIG_MYCLASS) // declare last for better alignment and if cast
 *     }
 *
 *     // DECL_TYPESIG_AUTO() is a shorthand that expands a constructor
 *     // and destructor to automatically set the type signature.
 *     // Using it in POD structs is recommended but may be inappropriate in complex classes.
 *     struct Pod  // Plain Old Data
 *     {
 *         int   data;
 *         DECL_TYPESIG_AUTO(Pod,TYPESIG_POD)  // !!!DECLARE LAST IN CASE OF CASTING!!!
 *     };
 *
 *     // Example of checking a type signature.
 *     // NOTE: For terseness, CHECK_TYPESIG() asserts that the pointer isn't NULL
 *     //       to obviate writing ASSERT(p) prior to CHECK_TYPESIG(p,TYPESIG_MYSTRUCT).
 *     //       If a pointer is allowed to be NULL, use CHECK_TYPESIG_NULL_OK() instead.
 *     void Do( Pod* pod )
 *     {
 *         CHECK_TYPESIG(pod,TYPESIG_POD);         // asserts pointer != NULL
 *         CHECK_TYPESIG_NULL_OK(pod,TYPESIG_POD); // alternative that allows NULL
 *     }
 *
 * @endverbatim
 *//*
 * LEGAL:	COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef DOXYGEN  // doxygen can't parse this correctly
#ifndef BASE_TYPESIG_HH
#define BASE_TYPESIG_HH 1

namespace base {

/// COMPILE_TYPESIGS controls compiling type signatures.
#ifndef COMPILE_TYPESIGS
#	if DEBUG
#		define COMPILE_TYPESIGS 1
#	else
#		define COMPILE_TYPESIGS 0
#   endif
#endif

/// Macro to declare a type signature as a member in a class/struct.
/// GetTypesig() is for compatibility with objects encapsulated inside shptr/SafePtrs
/// without having to use their PTR().
/// GetTypesig##TYPESIG() expands unambiguous name
/// in case a derivative has its own different typesig value.
#if COMPILE_TYPESIGS
#define DECL_TYPESIG( TYPESIG ) typesig_t typesig##TYPESIG; typesig_t GetTypesig##TYPESIG( void ) const { return typesig##TYPESIG; }
#else
#define DECL_TYPESIG( TYPESIG )
#endif

/// Macro to declare a type signature member and expand
/// a constructor/destructor to automatically set it.
/// Recommended for structs but isn't appropriate in most classes.
#define DECL_TYPESIG_AUTO( CLASS, TYPESIG )           \
	CLASS()  {        SET_TYPESIG( this, TYPESIG ); } \
	~CLASS() { INVALIDATE_TYPESIG( this, TYPESIG ); } \
	DECL_TYPESIG( TYPESIG )

/// Macro to assign a type signature member.
/// Ensure typesig value isn't greater than what typesig member can hold.
#if COMPILE_TYPESIGS
#define SET_TYPESIG( OBJ, TYPESIG ) (OBJ)->typesig##TYPESIG = TYPESIG;
#else
#define SET_TYPESIG( OBJ, TYPESIG )
#endif

/// Macro to check a type signature -- WILL FAIL IF PASSED A NULL POINTER!!
#if COMPILE_TYPESIGS
#define CHECK_TYPESIG( OBJ, TYPESIG ) ASSERT( (OBJ)->GetTypesig##TYPESIG() == TYPESIG )
#else
#define CHECK_TYPESIG( OBJ, TYPESIG )
#endif

/// Macro to check a type signature -- will accept NULL pointer (ok)
#if COMPILE_TYPESIGS
#define CHECK_TYPESIG_NULL_OK( OBJ, TYPESIG ) ASSERT( ((OBJ) == NULL) || ((OBJ)->GetTypesig##TYPESIG() == TYPESIG) )
#else
#define CHECK_TYPESIG_NULL_OK( OBJ, TYPESIG )
#endif

/// Macro to invalidate a type signature member.
/// First check if it was valid in the first place
#if COMPILE_TYPESIGS
#define INVALIDATE_TYPESIG( OBJ, TYPESIG ) ( CHECK_TYPESIG((OBJ),TYPESIG), ((OBJ)->typesig##TYPESIG = TYPESIG_INVALID) )
#else
#define INVALIDATE_TYPESIG( OBJ, TYPESIG )
#endif

/// Macro to return true if a type signature is correct (doesn't assert).
#if COMPILE_TYPESIGS
#define TYPESIG_OK( OBJ, TYPESIG ) ( ((OBJ) != NULL) && ((OBJ)->GetTypesig##TYPESIG() == TYPESIG) )
#else
#define TYPESIG_OK( OBJ, TYPESIG ) (true)
#endif

} // namespace base

#endif // BASE_TYPESIG_HH
#endif // DOXYGEN
