/*****************************************************************************
 * $LastChangedDate: 2008-09-20 10:29:31 -0400 (Sat, 20 Sep 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	control module configuration.
 *//*
 * LEGAL:	COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define CONTROL_CONF_CC 1
#include "base/module.hh"
#include "base/misc.hh"
#include "base/conf.hh"
#include "base/conf_txt.hh"
using namespace base;
#include "control/module.hh"
#include "control/conf.hh"
#include "control/joystick.hh"
#include "control/defs.hh"

namespace control {

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  Conf  ///////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Conf::Conf( void )
:	mValid(false),
	// cmd-line args
	mJoyEnabled(false),      // -j arg
	mJoyAxis3(false),        // a -j arg
	mJoyCalibration(false),  // a -j arg
	mJoystickConfFilename("joystick.conf"),
	// .conf file: polling freq
	mJoystickTickFreq(defs::DEFAULT_JOYSTICK_TICK_FREQ)
{
}

/*****************************************************************************
 * Initialize/read configuration.
 *****************************************************************************/
void
Conf::Init( int argc, char** argv, const string& confTxtFilename )
{
BEGIN_ONCE {

	mValid = true;  // now conf is valid

	ParseCmdlineArgs( argc, argv );   // cmdline args parsed prior to reading .conf file
	ReadConfTxt( confTxtFilename );  // read from .conf file
	ReadJoystickConfTxt();
	ParseCmdlineArgs2( argc, argv );  // cmdline args override .conf file

} END_ONCE
}

// Read joystick configuration.
void
Conf::ReadJoystickConfTxt( void )
{
	try
	{
		ModuleConfTxt confTxt;
		// Handle if -jc cmd-line arg was passed (don't use data dir).
		// Allows the user to set configuration for a specific joystick.
		if ( mJoystickConfFilename == "joystick.conf" )
		{
			confTxt.Read( string(GetDataDir()) + string("joystick.conf") );
		}
		else
		{
			CLOG << "Using joystick conf " << mJoystickConfFilename << endl;
			confTxt.Read( mJoystickConfFilename );  // -jc was passed on cmdline
		}
	}
	catch (...)
	{
		CERROR << "ERROR: cannot read joystick.conf" << endl;
		EXIT( 1 );
	}
}

/*****************************************************************************
 * Parse cmdline args.
 *****************************************************************************/
void
Conf::ParseCmdlineArgs( int argc, char** argv )
{
// Too early to enable joystick.
#define JOY_ARG() {{ mJoyEnabled = true; }}

	// Parse args.
	for ( int i = 1; i < argc; ++i )
	{
		// -j (enable joystick)?
		if ( STREQ( argv[i], "-j" ) )
		{
			JOY_ARG();
			CmdargDone( i );
		}
		// Print joystick values (to help user to calibrate)?
		else if ( STREQ( argv[i], "-jp" ) )
		{
			// -jp imples enabling joystick but higher-level code now
			// also parses -jp and shows joystick values in GUI (used to be cout).
			JOY_ARG();
			CmdargDone( i );
		}
		else if ( STREQ( argv[i], "-jc" ) )
		{
			JOY_ARG();
			mJoystickConfFilename = Argstr( i+1, argc, argv );
			CmdargDone( i, 2 );
			i += 1;
		}
#if OS_LINUX || OS_FREEBSD
		// Switch to Linux joystick driver (FreeBSD has a port that can emulate it).
		else if ( STREQ(argv[i],"-jlinux") )
		{
			JOY_ARG();
			input::Joystick::GetInstance().SetJoystickDriver( input::Joystick::eJoystickDriver_LINUX );
			CmdargDone( i );
		}
#endif
#if COMPILE_GLUT_JOYSTICK
		// Switch to GLUT joystick driver.
		else if ( STREQ(argv[i],"-jglut") )
		{
			JOY_ARG();
			input::Joystick::GetInstance().SetJoystickDriver( input::Joystick::eJoystickDriver_GLUT );
			CmdargDone( i );
		}
#endif

	}
}

void
Conf::ParseCmdlineArgs2( int argc, char** argv )
{
	// NOP
}

/*****************************************************************************
 * Read from .conf file.
 *****************************************************************************/
void
Conf::ReadConfTxt( const string& confTxtFilename )
{
	try
	{
		ModuleConfTxt confTxt;
		confTxt.Read( string(GetDataDir()) + string(confTxtFilename) );
	}
	catch (...)
	{
		CERROR << "ERROR: cannot read " << confTxtFilename << endl;
		EXIT( 1 );
	}
}

void
Conf::ModuleConfTxt::ProcessField( string& field, ifstream& conf )
{
	bool error = false;

	control::Joystick& joystick = control::Joystick::GetInstance();

	// -----------------------  polling freq  ----------------------------------
	if ( field == "JoystickTickFreq" )
	{
		int i; conf >> i;
		CONTROL_CONF.mJoystickTickFreq = Milliseconds(i);
		if ( i < 1 || i >= 1000 ) error = true;
	}
	// -----------------------  joystick.conf  -----------------------------------------
	else if ( field == "PitchAxis" )
	{
		int i; conf >> i;
		joystick.MapAxis( control::Joystick::AxisMap::PITCH, i );
	}
	else if ( field == "RollAxis" )
	{
		int i; conf >> i;
		joystick.MapAxis( control::Joystick::AxisMap::ROLL, i );
	}
	else if ( field == "YawAxis" )
	{
		int i; conf >> i;
		joystick.MapAxis( control::Joystick::AxisMap::YAW, i );
	}
	else if ( field == "ThrottleAxis" )
	{
		int i; conf >> i;
		joystick.MapAxis( control::Joystick::AxisMap::THROTTLE, i );
	}
	else if ( field == "PitchRange" )
	{
		// Example:
		// PitchRange -32767 -1000 1000 32767
		control::Joystick::Calibration calibration = joystick.GetCalibration();
		conf >> calibration.mPitchUp.mMax;  // max,min,min,max
		conf >> calibration.mPitchUp.mMin;
		conf >> calibration.mPitchDown.mMin;
		conf >> calibration.mPitchDown.mMax;
		joystick.SetCalibration( calibration );
	}
	else if ( field == "RollRange" )
	{
		control::Joystick::Calibration calibration = joystick.GetCalibration();
		conf >> calibration.mRollLeft.mMax;
		conf >> calibration.mRollLeft.mMin;
		conf >> calibration.mRollRight.mMin;
		conf >> calibration.mRollRight.mMax;
		joystick.SetCalibration( calibration );
	}
	else if ( field == "YawRange" )
	{
		control::Joystick::Calibration calibration = joystick.GetCalibration();
		conf >> calibration.mYawLeft.mMax;
		conf >> calibration.mYawLeft.mMin;
		conf >> calibration.mYawRight.mMin;
		conf >> calibration.mYawRight.mMax;
		joystick.SetCalibration( calibration );
	}
	else if ( field == "ThrottleRange" )
	{
		// Throttle is different as it has one range:
		// ThrottleRange  -32767 32767
		control::Joystick::Calibration calibration = joystick.GetCalibration();
		conf >> calibration.mThrottle.mMin;
		conf >> calibration.mThrottle.mMax;
		joystick.SetCalibration( calibration );
	}

    // Error?
	if ( error )
	{
		CERROR << "ERROR[control conf]: parameter for " << field << " is invalid." << endl;
		EXIT( 1 );
	}
}

/*****************************************************************************
 * Print configuration.
 *****************************************************************************/
ostream&
operator<<( ostream& strm, const Conf& conf )
{
	return strm;
}

} // namespace control
