/*****************************************************************************
 * $LastChangedDate: 2008-08-27 19:45:17 -0400 (Wed, 27 Aug 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Higher-level Joystick class.
 *//*
 * LEGAL:	COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef CONTROL_JOYSTICK_HH
#define CONTROL_JOYSTICK_HH 1

#define GET_CONTROL_JOYSTICK() (control::Joystick::GetInstance())

#include "base/event.hh"
using namespace base;
#include "input/module.hh"
#include "input/joystick.hh"
using namespace input;

namespace control {

////////////////////////////////////////////////////////////////////////////////
/// @brief Joystick control of a Craft, joystick calibration.
///
/// This is the client of the lower-level input::Joystick.
/// This handles axis events: the current Craft is rotated by axis events.
/// This doesn't handle button events, rather, it defers them to a listener.
///
/// This class cannot be derived from input::Joystick because of
/// conflicting members (mEvent) and input::Joystick is a Singleton.
/// Instead, this class is composed with (built over) input::Joystick.
///
/// To register a button handler, register an event listener
/// with Joystick::mEventButton (see base/event.hh).
///
/// Event classes:
/// base::Event is an object for broadcasting to listeners.
/// Joystick::Event is just a simple struct.
///
class Joystick
{
PREVENT_COPYING(Joystick)
public:
	CLASS_CONST int MAX_AXIS   = 4;   // pitch,roll,yaw,throttle = 4
	CLASS_CONST int MAX_BUTTON = 10;  // arbitrary

public:
	/// @brief Defines a range of axis values for calibration.
	/// Roll/pitch/yaw axises have two ranges with a deadzone in between.
	/// Throttle axis has one range.
	///
	struct Calibration
	{
		/// @brief min,max
		///
		struct AxisRange
		{
			AxisRange( void )
			: mMin(0), mMax(input::Joystick::MAX_AXIS_VALUE)
			{ }
			int mMin, mMax;  // NOTE: min/max may be both positive or negative
		};
		AxisRange mPitchUp,  mPitchDown;
		AxisRange mRollLeft, mRollRight;
		AxisRange mYawLeft,  mYawRight;
		AxisRange mThrottle;
	};

private:
	/// Pitch/roll/yaw have two active segments/ranges with a deadzone between.
	/// Throttle has one segment/range.
    enum eMove
    {
    	eMove_AT_CENTER,	// no movement
    	eMove_PITCH_UP,
    	eMove_PITCH_DOWN,
    	eMove_ROLL_LEFT,
    	eMove_ROLL_RIGHT,
    	eMove_YAW_LEFT,
    	eMove_YAW_RIGHT,
    	eMove_THROTTLE,		// throtle has no deadzone/center
    };

public:
	/// @brief Remaps joystick axises and indicates which axis had an event.
	/// Calling the If*() methods effectively remaps the axises
	/// (that is, remapping isn't done by changing the Event struct).
	///
	class AxisMap
	{
	friend class Joystick;
	public:
		enum { ROLL, PITCH, YAW, THROTTLE };
		AxisMap( void );
		bool	IfRoll( const input::Joystick::Event& event );
		bool	IfPitch( const input::Joystick::Event& event );
		bool	IfYaw( const input::Joystick::Event& event );
		bool	IfThrottle( const input::Joystick::Event& event );
	private:
		vector<int> mAxises;
	};
	friend class AxisMap;

private:
	/// @brief Polls joystick events.
	///
	class TimerFunctor : public Timer::Functor
	{
	public:
				TimerFunctor( control::Joystick& joystick ) : mJoystick(joystick) { }
		void	operator()( const Milliseconds millisecElapsed );
		control::Joystick&  mJoystick;
	};
	friend class TimerFunctor;

private:
    Joystick( void );  // Singleton
    ~Joystick();

public:
    CLASS_METHOD control::Joystick& GetInstance( void )  // Singleton
	{
	    if ( UX( msInstance == NULL ) )
	        msInstance = new Joystick;
	    return *msInstance;
	}
	bool				IfEnabled( void );
	bool				Enable( bool enable = true );
	void				SetCalibration( const Calibration& calibration );
	Calibration			GetCalibration( void );
	void				MapAxis( const int axisLogical, const int axisReal );
	void				EnableAutofire( bool f );
	int                 GetAxisValue( uint axis );

private:
	void                EnableInitial( void );
	void                EnableSubsequent( bool enable );
	void                Poll( void );
	void				ProcessAxisEvent( const input::Joystick::Event& event );
	void				ProcessButtonEvent( const input::Joystick::Event& event );
	fp					ComputeFraction( const input::Joystick::Event& event, const uint axis, const Calibration::AxisRange& axisRange, const int dir );

public:
	base::Event<input::Joystick::Event> mEventButton;  //< for broadcasting button on/off event to registered listeners

private:
    CLASS_VAR Joystick*		msInstance;
	bool					mEnableInitial;		///< if Enable(true) was called (regardless if it succeeded or not)
	Calibration				mCalibration;		///< defines range of axis values
	AxisMap					mAxisMap;			///< to reverse an axis
	input::Joystick::Event	mEvent;				///< current event
	bool					mAutofire;
    TinyArray<MAX_AXIS,int> mAxisValues;        ///< for calibration (not used internally)
};

} // namespace control

#endif // CONTROL_JOYSTICK_HH
