/*****************************************************************************
 * $LastChangedDate: 2008-12-03 22:27:51 -0500 (Wed, 03 Dec 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Small particles and particle-system.
 * @remarks This implementation was kept simple.  For now, decided to omit
 *          implementing a particle updating mechanism and per-particle attributes.
 *//*
 * LEGAL:	COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define FX_PARTSYS_SMALL_CC 1
#include "base/module.hh"
using namespace base;
#include "object/module.hh"
using namespace object;
#include "world/module.hh"
using namespace world;
#include "fx/module.hh"
#include "fx/partsys_small.hh"

#include <osg/Point>
#include <osg/PointSprite>

namespace fx {

INTERN const uint SMALL_PARTICLES_LIMIT   = 5000;
INTERN const uint SMALL_PARTICLES_RESERVE = 5000;

////////////////////////////////////////////////////////////////////////////////
/////////////////////////  SmallParticleDrawable  //////////////////////////////
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
/// @brief This is an OSG Drawable object that directly renders point-sprites.
///
/// The reason for directly-rendering is to that the size of every point-sprite
/// needs to be computed individually.  Issuing glPointSize() for every particle
/// is faster and less-cumbersome than composing nodes with various point sizes.
///
/// When OSG has called drawImplementation(), OSG has set OpenGL state
/// for rendering OpenGL "point sprites".
///
class SmallParticleDrawable : public osg::Drawable
{
private:
    /*****************************************************************************
     * ctor subroutine.
     *****************************************************************************/
    void Init( void )
    {
        SET_TYPESIG(this,TYPESIG_SMALL_PARTICLE_DRAWABLE);

        // Pre-allocate.
        mVertexArray.reserve( SMALL_PARTICLES_RESERVE );

        // Disable display list else only the OpenGL calls
        // of the first particle would be captured.
        setUseDisplayList( false );
    }

public:
    META_Object(Palomino,SmallParticleDrawable)  // OSG macro

    /*****************************************************************************
     * ctor/dtor.
     * @param   color
     *          Applied to all particles.
     * @param   range
     *          Distance in world space where particle is scaled-down to smallest size.
     * @param   minWidth, maxWidth
     *          Limits of point size scaled by distance of their vertex from viewpoint.
     *          This simulates 3D:2D projection.
     *****************************************************************************/
    SmallParticleDrawable( const RGBA color, const fp range,  const fp minWidth, const fp maxWidth )
    :   osg::Drawable(),
        mColor(color),
        mRange(range),
        mRangeInv(1.0f / range),
        mMinWidth(minWidth),
        mMaxWidth(maxWidth),
        mVertexArray(),
        mBoundingBox()
    {
        Init();
    }

    // (for OSG)
    SmallParticleDrawable( void )
    :   osg::Drawable(),
        mColor(),
        mRange(1000.0f),
        mRangeInv(1000.0f),
        mMinWidth(1.0f),
        mMaxWidth(64.0f),
        mVertexArray(),
        mBoundingBox()
    {
    }

    // (unexpected)
    SmallParticleDrawable( const SmallParticleDrawable& other, const osg::CopyOp& copyop = osg::CopyOp::SHALLOW_COPY)
    :   osg::Drawable(),
        mColor(other.mColor),
        mRange(other.mRange),
        mRangeInv(other.mRangeInv),
        mMinWidth(other.mMinWidth),
        mMaxWidth(other.mMaxWidth),
        mVertexArray(other.mVertexArray),  // TOO SLOW
        mBoundingBox(other.mBoundingBox)
    {
    ASSERT(false);  // copying vertex array would be too SLOW, catch if copy ctor is ever called
        Init();
    }

    virtual ~SmallParticleDrawable()
    {
        INVALIDATE_TYPESIG(this,TYPESIG_SMALL_PARTICLE_DRAWABLE);
    }

    /*****************************************************************************
     * Draw array of small particles.
     *****************************************************************************/
	virtual void drawImplementation( osg::RenderInfo& renderInfo ) const
    {
    CHECK_TYPESIG(this,TYPESIG_SMALL_PARTICLE_DRAWABLE);

        const uint vcnt = mVertexArray.size();
        if ( UX( vcnt == 0 ) )
            return;

        // The inverse of view matrix has the correct eye position.
        const osg::Vec3  eye  = renderInfo.getCurrentCamera()->getInverseViewMatrix().getTrans();
        const osg::Vec3* v    = &mVertexArray[0];     // first vertex
      //const osg::Vec3* vEnd = &mVertexArray[vcnt];  // wrong: operator[] will throw exception!
        const osg::Vec3* vEnd = v + vcnt;             // end vertex (end as in STL)

        // Save current color.
        GLfloat savedColor[4];
        glGetFloatv( GL_CURRENT_COLOR, savedColor );

        // Set color.
        GLfloat color[4];
        mColor.Get4f<GLfloat>( color );  // RGBA object --> color[4]
        glColor4fv( color );

        // For each vertex.
        for ( ; v < vEnd; ++v )
        {
        ASSERT( not mVertexArray.empty() );

            const fp dist      = Min<fp>( mRange, Distance(eye,*v) );
            const fp ratio     = (mRange - dist) * mRangeInv;
            const fp pointSize = Max<fp>( mMinWidth, std::pow(ratio,4.0f) * mMaxWidth );
            glPointSize( pointSize );
            glBegin( GL_POINTS );
                glVertex3fv( reinterpret_cast<const GLfloat*>( v ) );
            glEnd();
        }

        // Restore color.
        glColor4fv( savedColor );
    }

    /*****************************************************************************
     * Compute bound of small particles.
     * Which is the fastest?
     * Bounding all point-sprites in one box or dividing them into multiple boxes (?).
     *****************************************************************************/
    osg::BoundingBox computeBound() const
    {
    CHECK_TYPESIG(this,TYPESIG_SMALL_PARTICLE_DRAWABLE);

        return mBoundingBox;
    }

    /*****************************************************************************
     * Add particle.
     *****************************************************************************/
    void
    AttachParticle( const osg::Vec3& v )
    {
    CHECK_TYPESIG(this,TYPESIG_SMALL_PARTICLE_DRAWABLE);

        // Too many particles will slow FPS.
        if ( EX( mVertexArray.size() < SMALL_PARTICLES_LIMIT ) )
        {
            // Append vertex.
            mVertexArray.push_back( v );

            // Expand bounding box.
            mBoundingBox.expandBy( v );

            // Force recomputing bounding box.
            dirtyBound();
        }
    }

    /*****************************************************************************
     * @return Vertex count (amount of particles).
     *****************************************************************************/
    uint
    GetVertexCnt( void ) const
    {
    CHECK_TYPESIG(this,TYPESIG_SMALL_PARTICLE_DRAWABLE);

        return mVertexArray.size();
    }

    /*****************************************************************************
     * Set color of all particles.
     *****************************************************************************/
    void
    SetColorAll( const RGBA color )
    {
    CHECK_TYPESIG(this,TYPESIG_SMALL_PARTICLE_DRAWABLE);

        mColor = color;
    }

private:
    RGBA                mColor;         ///< color of all point sprites
    const fp            mRange;         ///< width of point-sprite is minimized at limit of range
    const fp            mRangeInv;
    const fp            mMinWidth;      ///< minimum size of point-sprites
    const fp            mMaxWidth;      ///< maximum size of point-sprites
    vector<osg::Vec3>   mVertexArray;   ///< not osg::Vec3Array which lacks swap()
    osg::BoundingBox    mBoundingBox;   ///< no need for ref_ptr (isn't derived from osg::Referenced)
    DECL_TYPESIG(TYPESIG_SMALL_PARTICLE_DRAWABLE);
};

////////////////////////////////////////////////////////////////////////////////
///////////////////////  SmallParticleSystem::Args  ////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Args for SmallParticleSystem ctor.
 * @param   stateSet
 *          Texture can be specified in stateset.
 * @param   stateBits
 *          For state-sorting (tells World which node to place this under).
 * @param   color
 *          Of all particles.
 * @param   rangeMin, rangeMax
 *          Range in 3D space when 2D point sprites are scaled to min,max sizes.
 *          This simulates a 3D appearance.
 * @param   minWidth, maxWidth
 *          min,max of width of point sprites (limited by hardware, 64 pixels max is typical).
 * Example:
 * new SmallParticleSystem( SmallParticleSystem::Args( Texture::NewTextureAsStateSet( "smoke.png" ) ) );
 *****************************************************************************/
SmallParticleSystem::Args::Args( const Milliseconds lifetime,
                                 const bool timerEnabled,
                                 RefPtr<osg::StateSet> stateSet,
                                 const StateBits stateBits,
                                 const RGBA color,
                                 const Meter geoRange,
                                 const fp minWidth,
                                 const fp maxWidth )
:   ParticleSystem::Args(lifetime,timerEnabled),
    mStateSet(stateSet),
    mStateBits(stateBits),
    mColor(color),
    mRange(Geo::Meters2Sim(geoRange)),
    mMinWidth(minWidth),
    mMaxWidth(maxWidth)
{
    // NOP
}

////////////////////////////////////////////////////////////////////////////////
///////////////////////////  SmallParticleSystem  //////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
SmallParticleSystem::SmallParticleSystem( const Args& args )
:   ParticleSystem(args),
    // OSG components:
    mAttachedToSceneGraph(false),
    mStateSet(args.mStateSet),
    mStateBits(args.mStateBits),
    mGeode(new osg::Geode),
    mGeom(new osg::Geometry),
    mSmallParticleDrawable(new SmallParticleDrawable(args.mColor,args.mRange,args.mMinWidth,args.mMaxWidth))
{
    SET_TYPESIG(this,TYPESIG_SMALL_PARTICLE_SYSTEM);

    // Compose OSG components:
    // SmallParticleDrawable is the reason why some node components are omitted.
    mGeode->setStateSet( mStateSet.get() );
    mGeode->addDrawable( mSmallParticleDrawable.get() );

    // Configure point sprite:
    mStateSet->setAttribute( new osg::Point );
    mStateSet->setTextureAttributeAndModes( 0, new osg::PointSprite, osg::StateAttribute::ON );

    // (The caller should have passed a stateset with a texture attached to it)
}

SmallParticleSystem::~SmallParticleSystem()
{
    // Detach from scene-graph.
    if ( mAttachedToSceneGraph )
    {
        GET_WORLD().Detach( mGeode.get() );
    }

    INVALIDATE_TYPESIG(this,TYPESIG_SMALL_PARTICLE_SYSTEM);
}

/*****************************************************************************
 * Append SmallParticle.
 * Time-of-death is absoluete (add current time to lifetime).
 *****************************************************************************/
void
SmallParticleSystem::AttachParticle( const SmallParticle& smallParticle )
{
CHECK_TYPESIG(this,TYPESIG_SMALL_PARTICLE_SYSTEM);

    // Lazy-attach to scene-graph.
    if ( UX( not mAttachedToSceneGraph ) )
    {
        mAttachedToSceneGraph = true;
        GET_WORLD().Attach( mGeode.get(), mStateBits, mStateSet );
    }

    // Add vertex to Drawable.
    mSmallParticleDrawable->AttachParticle( smallParticle.mPosition );
}

/*****************************************************************************
 * @return Amount of particles.
 *****************************************************************************/
uint
SmallParticleSystem::GetParticleCnt( void ) const
{
CHECK_TYPESIG(this,TYPESIG_SMALL_PARTICLE_SYSTEM);

    return mSmallParticleDrawable->GetVertexCnt();
}

/*****************************************************************************
 * Set color of all particles.
 *****************************************************************************/
void
SmallParticleSystem::SetColorAll( const RGBA color )
{
CHECK_TYPESIG(this,TYPESIG_SMALL_PARTICLE_SYSTEM);

    return mSmallParticleDrawable->SetColorAll( color );
}

} // namespace fx
