/*****************************************************************************
 * $LastChangedDate: 2008-11-27 18:22:19 -0500 (Thu, 27 Nov 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Small particle-system.
 *//*
 * LEGAL:	COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef FX_PARTSYS_SMALL_HH
#define FX_PARTSYS_SMALL_HH 1

#include "base/timer.hh"
#include "fx/partsys.hh"

namespace fx {

class SmallParticleDrawable;

////////////////////////////////////////////////////////////////////////////////
/// @brief Opaque struct that defines a small particle.
///
struct SmallParticle
{
    WorldVertex     mPosition;
};

////////////////////////////////////////////////////////////////////////////////
/// @brief Particle-system based on OpenGL point-sprites.
///
/// This particle system is very fast but it has deficiencies.
/// Left unimplemented was a mechanism to update particles.
/// SmallParticleSystem has a crude way to adjust size of point sprites
/// by distance (but it's imperfect and hardware limits point size).
///
/// Texture is specified by stateset.
/// Example:
/// new SmallParticleSystem( SmallParticleSystem::Args( Texture::NewTextureAsStateSet( "smoke.png" ) ) );
///
class SmallParticleSystem : public ParticleSystem
{
PREVENT_COPYING(SmallParticleSystem)
typedef ParticleSystem Parent;
friend class FX; friend class shptr<SmallParticleSystem>;
friend class SmallParticleDrawable;
public:
    /// @brief Args for SmallParticleSystem ctor.
    struct Args : public ParticleSystem::Args
    {
        Args( const Milliseconds    lifetime,                      // base
              const bool            timerEnabled = TIMER_ENABLED,  // base
              RefPtr<osg::StateSet> stateSet = new osg::StateSet,
              const StateBits       stateBits = ParticleSystem::DefaultStateBits(),
              const RGBA            color = fx::defs::DEFAULT_SMALL_PARTSYS_COLOR,
              const Meter           geoRange = fx::defs::DEFAULT_SMALL_PARTSYS_GEO_RANGE,
              const fp              minWidth = fx::defs::DEFAULT_SMALL_PARTSYS_MIN_WIDTH,
              const fp              maxWidth = fx::defs::DEFAULT_SMALL_PARTSYS_MAX_WIDTH );

        RefPtr<osg::StateSet>   mStateSet;      ///< can specify texture
        const StateBits         mStateBits;     ///< for state-sorting
        const RGBA              mColor;         ///< color of all particles
        const fp                mRange;         ///< particle reaches it smallest 2D width at this 3D distance
        const fp                mMinWidth, mMaxWidth;
    };

public:
                    SmallParticleSystem( const Args& arg );
                    ~SmallParticleSystem();
    void            AttachParticle( const SmallParticle& smallParticle );
    uint            GetParticleCnt( void ) const;
    void            SetColorAll( const RGBA color );

private:
    // OSG components:
    bool                            mAttachedToSceneGraph;      ///< lazy-attached in case of no particles
    RefPtr<osg::StateSet>           mStateSet;                  ///< specifies texture (and any other state)
    StateBits                       mStateBits;                 ///< for state-sorting
    RefPtr<osg::Geode>              mGeode;
    RefPtr<osg::Geometry>           mGeom;
    RefPtr<SmallParticleDrawable>   mSmallParticleDrawable;     ///< for adjusting size of every point sprite
public:
	DECL_TYPESIG(TYPESIG_SMALL_PARTICLE_SYSTEM)
};

} // namespace fx

#endif // FX_PARTSYS_SMALL_HH
