/*****************************************************************************
 * $LastChangedDate: 2008-12-05 22:19:57 -0500 (Fri, 05 Dec 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Light class.
 * @verbatim
 *
 * Issues/solutions with lighting in OSG:
 * --------------------------------------
 * [2008/06 applies to OSG 2.4.0 ... 2.5.x]
 *
 * The first problem is that OSG allows a child node to override the
 * SAME KIND of state set by its parent.  This means that, in Palomino,
 * if a particular 3D model has its own osg::LightSource, the change
 * in the light source will affect other nodes in the scene graph!
 * So if darker-more-specular lighting of the SR-71 Blackbird
 * will effectively spill over to other nodes (F-18 Hornet).
 *
 *      LightSrc[0]
 *        /    \
 *     SR-71    F-18
 *  LightSrc[0]
 *
 * The second problem is unintentional transformation of child nodes
 * by osg::LightGroup.  This occurs if osg::LightSource is configured
 * as RELATIVE_RF and osg::LightSrc::SetPosition() is called.
 * OSG documents that the modelview matrix is applied for RELATIVE_RF
 * which has the side-effect of transforming its children
 * (not sure if this side-effect is intentional or not).
 * The solution is to set ABSOLUTE_RF (which inhibits transformation by OSG)
 * then register a pre-draw camera callback to manually transform
 * the light's position from world space to eye space.
 * BTW, this transformation is more than a workaround,
 * as shaders (if implemented) will need the app to do this transformation.
 *
 * @endverbatim
 *//*
 * LEGAL:   COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define GRAPH_LIGHT_CC 1
#include "base/module.hh"
using namespace base;
#include "gfx/module.hh"
#include "gfx/osg.hh"
using namespace gfx;
#include "view/module.hh"
#include "view/events.hh"
using namespace view;
#include "graph/module.hh"
#include "graph/light.hh"

#include <osg/LightModel>

namespace graph {

////////////////////////////////////////////////////////////////////////////////
/// @brief Transforms the light's position from world to eye space.
///
CLASS_EVENT_LISTENER( LightPreDrawViewListener, shptr<View>, Light, mLight )

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////  Light  ///////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

osg::Vec4 Light::msDefaultBaseAmbient( 0.0f, 0.0f, 0.0f, 0.0f );
fp        Light::msMulColor = 1.0f;

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Light::Light( const uint lightNum,
              const WorldVertex& lightPos,
              const RGBA lightColorAmbient,
              const RGBA lightColorDiffuse,
              const RGBA lightColorSpecular )
:	mLightNum(lightNum),
	mLight(new osg::Light),
	mLightSrcGroup(new osg::LightSource),
	mWorldPos(lightPos),
	mPreDrawViewListener(new LightPreDrawViewListener(*this))
{
	SET_TYPESIG(this,TYPESIG_LIGHT);

	// Configure osg::Light (non-node).
    mLight->setLightNum( lightNum );
    mLight->setAmbient(  RGBA2VEC4F( lightColorAmbient ) );
    mLight->setDiffuse(  RGBA2VEC4F( lightColorDiffuse ) );
    mLight->setSpecular( RGBA2VEC4F( lightColorSpecular ) );
//  mLight->setPosition();  // delayed

	// Configure osg::LightSource (osg::Group derivative).
	// RELATIVE_RF (default) would be used but it causes incorrect
	// transformation of its children.  The alternative is to
	// transform the light's position during the pre-draw camera callback.
    mLightSrcGroup->setLight( mLight.get() );
	mLightSrcGroup->setReferenceFrame( osg::LightSource::ABSOLUTE_RF );

	// Override OpenGL's base ambient light.
	SetBaseAmbient( VEC4F2RGBA( msDefaultBaseAmbient ) );

	// Register functors to be called before rendering this View.
	EVENT_VIEW_PRE_DRAW_ONE_VIEW.Listen( mPreDrawViewListener.PTR() );
}

Light::~Light()
{
	// Unregister or crash into destroyed object!
	EVENT_VIEW_PRE_DRAW_ONE_VIEW.Unlisten( mPreDrawViewListener.PTR() );

	INVALIDATE_TYPESIG(this,TYPESIG_LIGHT);
}

/*****************************************************************************
 * Set position of Light (world space).
 *****************************************************************************/
void
Light::SetPosition( const WorldVertex& pos )
{
CHECK_TYPESIG(this,TYPESIG_LIGHT);

	mWorldPos = pos;
}

/*****************************************************************************
 * @return Position of Light (world space).
 *****************************************************************************/
WorldVertex
Light::GetPosition( void )
{
CHECK_TYPESIG(this,TYPESIG_LIGHT);

	return mWorldPos;
}

/*****************************************************************************
 * Set light colors.
 *****************************************************************************/
void
Light::SetAmbient( const RGBA color )
{
CHECK_TYPESIG(this,TYPESIG_LIGHT);

    mLight->setAmbient( MulColor(color) );
}

void
Light::SetDiffuse( const RGBA color )
{
CHECK_TYPESIG(this,TYPESIG_LIGHT);

    mLight->setDiffuse( MulColor(color) );
}

void
Light::SetSpecular( const RGBA color )
{
CHECK_TYPESIG(this,TYPESIG_LIGHT);

    mLight->setSpecular( MulColor(color) );
}

/*****************************************************************************
 * Configure OpenGL's global ambient light (base ambient).
 * This is rarely called.
 *****************************************************************************/
void
Light::SetBaseAmbient( const RGBA color )
{
	// Yes, if this is called again, redo.
    RefPtr<osg::LightModel> lightModel = new osg::LightModel;
    lightModel->setAmbientIntensity( MulColor(color) );
	mLightSrcGroup->getOrCreateStateSet()->setAttribute( lightModel.get() );
}

void // CLASS_METHOD
Light::SetDefaultBaseAmbient( const RGBA color )
{
	msDefaultBaseAmbient = MulColor(color);
}

/*****************************************************************************
 * Attach a node to this osg::LightSource (group node).
 *****************************************************************************/
void
Light::AttachNode( RefPtr<Node> node )
{
CHECK_TYPESIG(this,TYPESIG_LIGHT);

	mLightSrcGroup->addChild( node.get() );
}

/*****************************************************************************
 * Multiply a color.
 * By scaling the var msMulCol, day/night colors are supported.
 *****************************************************************************/
void // CLASS_METHOD
Light::SetMulColor( const fp mulColor )
{
	// TODO/FIXME: Light needs to be extended to have a container of its instances,
	// then colors of instances are re-multiplied.
	// This allows dynamically changing light colors (time-of-day).

	msMulColor = mulColor;
}

osg::Vec4
Light::MulColor( const RGBA color )
{
	// Don't multiply alpha.
	return osg::Vec4f( color[RR] / 256.0f * msMulColor,
                       color[GG] / 256.0f * msMulColor,
                       color[BB] / 256.0f * msMulColor,
                       color[AA] / 256.0f );
}

/*****************************************************************************
 * Transforms the light's position from world to eye space.
 *****************************************************************************/
void
LightPreDrawViewListener::operator()( shptr<View> view )
{
CHECK_TYPESIG(&mLight,TYPESIG_LIGHT);

	// Reflect.
	mLight.PreDraw( view );
}

void
Light::PreDraw( shptr<View> view )
{
CHECK_TYPESIG(view,TYPESIG_VIEW);
CHECK_TYPESIG(this,TYPESIG_LIGHT);

	// SUBTLE: W coordinate = 1.0 sets a POSITIONAL light in OpenGL.
	const EyeVertex ev = view->TransformWorld2Eye( mWorldPos );
    mLight->setPosition( osg::Vec4( ev[XX], ev[YY], ev[ZZ], 1.0f ) );
}

} // namespace graph
