/*****************************************************************************
 * $LastChangedDate: 2008-12-05 10:58:31 -0500 (Fri, 05 Dec 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	HUD text objects.
 * @remarks
 * In OSG, multiple lines of text is drawn by attaching nodes of
 * osg::Text objects, contained in Geodes, that define different
 * 2D positions and strings.  These are abstracted by the HudText class.
 *
 * Arcane magic and kludges:
 * -------------------------
 * - Right-alignment:
 *   This is done using HUD_TEXT_FONT_WIDTH_RATIO which is a magic value
 *   obtained by experimentation.  It is specific to the "arial.ttf" font.
 *   Using another font while orienting at right would probably result
 *   in misalignment.  Though it isn't used for normal left-alignment.
 *
 *//*
 * LEGAL:   COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define HUD_HUD_TEXT_CC 1
#include "base/module.hh"
#include "base/conf.hh"
#include "base/timer.hh"
using namespace base;
#include "gfx/module.hh"
#include "gfx/osg.hh"
using namespace gfx;
#include "graph/module.hh"
using namespace graph;
#include "view/module.hh"
using namespace view;
#include "hud/module.hh"
#include "hud/hud_text.hh"

#include <osg/CameraNode>

namespace hud {

INTERN const fp			HUD_TEXT_COLOR_ALPHA = 1.0f;
INTERN const osg::Vec4f	HUD_TEXT_COLOR( 0.8f, 1.0f, 0.8f, HUD_TEXT_COLOR_ALPHA );
INTERN const string		HUD_TEXT_FONT_NAME( "arial.ttf" );
INTERN const fp			HUD_TEXT_FONT_SIZE = 28.0f;
INTERN const fp         HUD_TEXT_FONT_WIDTH_RATIO  = 0.593f;  // [see issues] specific to "fonts/arial.ttf"
INTERN const fp         HUD_TEXT_FONT_HEIGHT_RATIO = 1.0f;

////////////////////////////////////////////////////////////////////////////////
//////////////////////////////  HudText::Args  /////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor.
 *****************************************************************************/
HudText::Args::Args( void )
:	mPos(0.0f,0.0f),
	mOffset(),
	mOrientation(eOrientation_BOTTOM_LEFT),
	mText(),
	mFontName(HUD_TEXT_FONT_NAME),
	mFontSize(HUD_TEXT_FONT_SIZE),
    mColor(HUD_TEXT_COLOR)
{
	mOffset[XX] = mOffset[YY] = 0;
}

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  HudText  ////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
HudText::HudText( const Args& args )
:	Parent(),
	mGeode(new osg::Geode),
	mOsgText(new osgText::Text),
	m3D(),
	mOffset(args.mOffset)
{
// Empty mText is ok.
ASSERT( args.mPos[XX] >= 0 and args.mPos[XX] < 50000 );
ASSERT( args.mPos[YY] >= 0 and args.mPos[YY] < 50000 );
ASSERT( args.mFontSize > 0 );
ASSERT( not args.mFontName.empty() );

	SET_TYPESIG(this,TYPESIG_HUD_TEXT);

	// Compose geode.
	mGeode->addDrawable( mOsgText.get() );

	// Set the orientation/alignment.
	const Vector2 pos = OrientPosition( args );
    mOsgText->setPosition( osg::Vec3( pos[XX], pos[YY], 0.0f ) );
	mOsgText->setText( args.mText );
    mOsgText->setFont( base::GetDataDir() + string("/fonts/") + args.mFontName );
    mOsgText->setCharacterSize( args.mFontSize, 1.0f );
	mOsgText->setColor( RGBA2VEC4F(args.mColor) );
}

HudText::~HudText()
{
	INVALIDATE_TYPESIG(this,TYPESIG_HUD_TEXT);
}

/*****************************************************************************
 * (private ctor subroutine)
 * Re-orient the position of a HudText.
 * The input position must be oriented at BOTTOM_LEFT.
 *****************************************************************************/
Vector2
HudText::OrientPosition( const Args& args )
{
CHECK_TYPESIG(this,TYPESIG_HUD_TEXT);

	Vector2 pos = args.mPos;

	// Estimate the width,height of the font.
	// KLUDGE: HUD_TEXT_FONT_WIDTH_RATIO is specific to to "arial.ttf".
	const fp fontWidth  = args.mFontSize * HUD_TEXT_FONT_WIDTH_RATIO;
	const fp fontHeight = args.mFontSize * HUD_TEXT_FONT_HEIGHT_RATIO;

	// Computing Y is straightforward.
	// Computing X for the right-alignment cases,
	// requires accounting for how many chars the text string has.
	switch ( args.mOrientation )
	{
		case eOrientation_BOTTOM_LEFT: default:
		{
			// NOP
		}
		break;

		case eOrientation_BOTTOM_RIGHT:
		{
			pos[XX] = HUD_ORTHO_X2 - pos[XX] - fontWidth * args.mText.size();
		}
		break;

		case eOrientation_TOP_LEFT:
		{
			pos[YY] = HUD_ORTHO_Y2 - pos[YY] - fontHeight;
		}
		break;

		case eOrientation_TOP_RIGHT:
		{
			pos[XX] = HUD_ORTHO_X2 - pos[XX] - fontWidth * args.mText.size();
			pos[YY] = HUD_ORTHO_Y2 - pos[YY] - fontHeight;
		}
		break;

		case eOrientation_CENTER:
		{
			pos[XX] = HUD_ORTHO_X2 * 0.5f - (fontWidth * args.mText.size() * 0.5f);
			pos[YY] = HUD_ORTHO_Y2 * 0.5f + pos[YY];  // client may add Y offset
		}
		break;
	}

	return pos;
}

/*****************************************************************************
 * Change the value of HudText.
 *****************************************************************************/
void
HudText::SetText( const char text[] )
{
CHECK_TYPESIG(this,TYPESIG_HUD_TEXT);

	mOsgText->setText( text );
}

void
HudText::SetText( const string& text )
{
CHECK_TYPESIG(this,TYPESIG_HUD_TEXT);

	mOsgText->setText( text.c_str() );
}

/*****************************************************************************
 * Set position by transforming 3D vertex into 2D.
 *****************************************************************************/
void
HudText::SetPosition3D( const WorldVertex& pos3 )
{
CHECK_TYPESIG(this,TYPESIG_HUD_TEXT);

	// Deferred to Update().
	m3D.mValid    = true;
	m3D.mPos3     = pos3;
	m3D.mLastTime = Milliseconds(0);
}

/*****************************************************************************
 * Set font size.
 *****************************************************************************/
void
HudText::SetFontSize( const fp fontSize )
{
CHECK_TYPESIG(this,TYPESIG_HUD_TEXT);

    mOsgText->setCharacterSize( fontSize, 1.0f );
}

/*****************************************************************************
 * Set color of text.
 *****************************************************************************/
void
HudText::SetColor( const RGBA rgba )
{
CHECK_TYPESIG(this,TYPESIG_HUD_TEXT);

	mOsgText->setColor( RGBA2VEC4F(rgba) );
}

/*****************************************************************************
 * Set drawing text with 4 gradient colors.
 *****************************************************************************/
void
HudText::SetGradientColors4( const TinyArray<4,RGBA>& colors4 )
{
CHECK_TYPESIG(this,TYPESIG_HUD_TEXT);
ASSERT( colors4.size() == 4 );

	mOsgText->setColorGradientCorners( RGBA2VEC4F(colors4[0]),
                                       RGBA2VEC4F(colors4[1]),
                                       RGBA2VEC4F(colors4[2]),
                                       RGBA2VEC4F(colors4[3]) );
	mOsgText->setColorGradientMode( osgText::Text::OVERALL );
}

/*****************************************************************************
 * Update callback.
 *****************************************************************************/
void
HudText::Update( View& view )
{
	// Deferred action of SetPosition3D().
	// Note that osgText::setPosition() is a bit slow,
	// so only call it a few times a second.
	const Milliseconds currTime = GET_TIMER().GetElapsedTime();
	if ( m3D.mValid
     and currTime - m3D.mLastTime > Milliseconds(defs::HUD_FREQ) )
	{
		// Remember time.
		m3D.mLastTime = currTime;

		// Don't draw if 3D vertex is behind viewpoint.
		const bool ifZFacing = view.IF_Z_FACING( m3D.mPos3 );
		mGeode->setNodeMask( ifZFacing ? ~0 : 0 );
		if ( ifZFacing )
		{
			// Transform 3D --> 2D.
			Vector2 pos2 = view.Project( m3D.mPos3 );

			// Remap 2D from viewport's dimensions to HUD's dimensions.
			const Rect<int> viewport = view.GetRect();
			pos2[XX] = Remap( pos2[XX], viewport.x2, HUD_ORTHO_X2 );
			pos2[YY] = Remap( pos2[YY], viewport.y2, HUD_ORTHO_Y2 );

			// Adjust so that text isn't over 3D object.
			pos2[XX] += mOffset[XX];
			pos2[YY] += mOffset[YY];

			// Directly change position of osg::Text.
		    mOsgText->setPosition( osg::Vec3( pos2[XX], pos2[YY], 0.0f ) );
		}
	}
}

} // namespace hud
