/*****************************************************************************
 * $LastChangedDate: 2008-09-30 16:12:50 -0400 (Tue, 30 Sep 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	HUD text objects.
 *//*
 * LEGAL:   COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef HUD_HUD_TEXT_HH
#define HUD_HUD_TEXT_HH 1

#include <osgText/Text>

namespace hud {

////////////////////////////////////////////////////////////////////////////
/// @brief Defines a string of text to be displayed in HUD.
///
/// Strings of text are displayed by the client allocating a HudText object.
/// HudText defines a 2D position and the value of the string.
/// HudText can be later modified or removed from the Hud.
///
/// HUD is auto-drawn by OSG since the HUD is comprised of a camera with
/// a 2D projection matrix and osg::Text objects which define a line of text.
///
class HudText : public HudObject
{
PREVENT_COPYING(HudText)
typedef HudObject Parent;
friend class Hud;
public:
	enum eOrientation
	{
		eOrientation_BOTTOM_LEFT,   // default
		eOrientation_BOTTOM_RIGHT,  // same order as OSG setColorGradientCorners()
		eOrientation_TOP_LEFT,
		eOrientation_TOP_RIGHT,
		eOrientation_CENTER,
	};

	// For setting RGBA array passed to SetGradientColors().
	enum eGradientColor
	{
		eGradientColor_TOP_LEFT,
        eGradientColor_BOTTOM_LEFT,
        eGradientColor_BOTTOM_RIGHT,
        eGradientColor_TOP_RIGHT,
	};

	/// @brief ctor args for HudText().
	///
	struct Args
	{
		Args( void );
		Vector2				mPos;
		TinyArray<2,int>	mOffset;
		eOrientation		mOrientation;
		string				mText;
		string				mFontName;
		fp					mFontSize;
		RGBA				mColor;
	};
					HudText( const Args& args );
					~HudText();
	void			SetText( const char text[] );
	void			SetText( const string& text );
	void			SetPosition3D( const WorldVertex& pos3 );
	void			SetFontSize( const fp fontSize );
	void			SetColor( const RGBA rgba );
	void			SetGradientColors4( const TinyArray<4,RGBA>& colors4 );
	RefPtr<Node>	GetNode( void ) { return mGeode.get(); }

private:
	Vector2			OrientPosition( const Args& args );
	virtual void	Update( View& view );

private:
	/// @brief Optional definition of 2D box in 3D terms.
	/// Becomes valid after SetPosition3D() is called.
	///
	struct Definition3D
	{
		Definition3D( void ) : mValid(false), mPos3(0,0,0), mLastTime(0) { }
		bool			mValid;
		WorldVertex		mPos3;
		Milliseconds	mLastTime;
	};

private:
	RefPtr<osg::Geode>		mGeode;		///< node
	RefPtr<osgText::Text>	mOsgText;	///< drawable (underlying OSG object)
	Definition3D			m3D;		///< enabled by SetPosition3D()
	TinyArray<2,int>		mOffset;
public:
	DECL_TYPESIG(TYPESIG_HUD_TEXT)
};

} // namespace hud

#endif // HUD_HUD_TEXT_HH
