/*****************************************************************************
 * $LastChangedDate: 2008-10-14 22:37:47 -0400 (Tue, 14 Oct 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Base low-level Joystick class (system-neutral).
 * @verbatim
 *
 * Axis values are returned in range -32767...+32767 as:
 *
 *           -32767
 *              |
 *    -32767  --o--  +3267
 *              |
 *           -32767
 *
 * @endverbatim
 *//*
 * LEGAL:	COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef INPUT_JOYSTICK_HH
#define INPUT_JOYSTICK_HH 1

namespace input {

class InputQueue;

////////////////////////////////////////////////////////////////////////////////
/// @brief Base Joystick class (system-neutral).
///
class Joystick
{
PREVENT_COPYING(Joystick)
friend class InputQueue;
public:
	/// Joystick axis extremes.
	CLASS_CONST int  MAX_AXIS_VALUE				= 0x7fff;				///< don't change
	CLASS_CONST uint DEFAULT_JOY_POLL_TICK_FREQ = DEFAULT_TICK_FREQ;	///< milliseconds, overriden by SetPollFreq()

    /// @brief Which joystick driver.
    enum eJoystickDriver
    {
        eJoystickDriver_NOP,
    #if OS_LINUX || OS_FREEBSD
        eJoystickDriver_LINUX,
    #endif
    #if COMPILE_GLUT_JOYSTICK  // not #elif
        eJoystickDriver_GLUT,
    #endif
    #if OS_WINDOWS
        eJoystickDriver_WINDOWS,
    #endif
    };

private:
	CLASS_CONST uint JOY_EVENT_ENQUEUE_LIMIT = 20;  ///< limit the enqueueing of events per tick

public:
	/// @brief Enumerations of axis and buttons are zero-based so -1 indicates none.
	enum
	{
		eAxis_NONE   = -1,
		eButton_NONE = -1,
	};

public:
	////////////////////////////////////////////////////////////////////////////////
	/// @brief Joystick event class.
	///
	/// Think of Event as a union: it should be assigned with
	/// either an axis or button event (not both).
	///
	class Event
	{
	COPYABLE(Event)
	public:
		Event( void )
		{
			Clear();
		}

		void Clear( void )
		{
			mAxis      = eAxis_NONE;
			mAxisVal   = 0;
			mButton    = eButton_NONE;
			mButtonVal = 0;
		}

		bool IfValid( void ) const
		{
			// Either.
			return (mAxis != eAxis_NONE) || (mButton != eButton_NONE);
		}

		bool IfEqualButtons( const Event& other ) const
		{
			return (mButton == other.mButton) and (mButtonVal == other.mButtonVal);
		}

	public:
		int		mAxis;			///< axis (0=X) or eAxis_NONE
		int		mAxisVal;		///< axis movement
		int		mButton;		///< button (0=first) or eButton_NONE
		bool	mButtonVal;		///< button on/off
	};

protected:
	// Only derivative can construct this.
					Joystick( void );
	virtual			~Joystick();

// Interface for client:
// SetPollFreq() can tune joystick (default value may or may not suffice).
public:
    CLASS_METHOD Joystick& GetInstance( void );  // returns system-specific derivative

	bool			Enable( bool flag = true );  ///< enable/disable joystick, returns true if present/enabled
	bool			IfEnabled( void ) { return mEnabled; }
    void            SetJoystickDriver( eJoystickDriver joystickDriver );
	virtual void	PrintStatus( void ) = 0;
	virtual void	SetPollFreq( const Milliseconds pollFreq );  // applicability depends on system

// Interface for InputQueue:
private:
	void			PollAndEnqueueEvents( const uint limit = JOY_EVENT_ENQUEUE_LIMIT );  ///< enqueues into InputQueue
protected:
	virtual bool	PollEvent( Event& event ) = 0;

// Template Method:
protected:
	virtual	bool	Open( void ) = 0;
	virtual void	Close( void ) = 0;

// Data members:
private:
    CLASS_VAR Joystick*     msInstance;
protected:
    CLASS_VAR int           msInstanceCnt;  ///< counts an instance of any class of Joystick
	bool                    mEnabled;		///< base Enable() assigns from derived Open()/Close()
	Milliseconds            mPollFreq;		///< driver-specific code may or may not have a use for this
};

} // namespace input

#endif // INPUT_JOYSTICK_HH
