/*****************************************************************************
 * $LastChangedDate: 2008-07-05 11:25:00 -0400 (Sat, 05 Jul 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Joystick device (Linux).
 * @remarks Documented at inp_joystick.hh
 *          This is compiled on FreeBSD also.
 *          FreeBSD's devel/linux-js port emulates a Linux joystick driver.
 *//*
 * LEGAL:	COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#if OS_LINUX || OS_FREEBSD

#define INPUT_JOYSTICK_LINUX_CC 1
#include "base/module.hh"
#include "base/conf.hh"
using namespace base;
#include "input/module.hh"
#include "input/joystick.hh"
#include "input/joystick_linux.hh"

namespace input {

////////////////////////////////////////////////////////////////////////////////
//////////////////////////  JoystickLinux  /////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
JoystickLinux::JoystickLinux( void )
:	mDeviceName("/dev/input/js0"),
	mDeviceName2("/dev/js0"),
	mFiledesc(-1)
{
    // Let base Joystick ctor catch more than once instance.
}

JoystickLinux::~JoystickLinux()
{
	// NOP
}

/*****************************************************************************
 * Open Linux joystick device.
 *****************************************************************************/
bool
JoystickLinux::Open( void )
{
	// Open the joystick device only once.  Repetively opening-and-closing
	// it causes superfluous events and unreliable joystick readings.
	if ( mFiledesc != -1 )
		return (mEnabled = true);  // even though redundant with Joystick::Open()

	// Open in non-blocking mode.
	mFiledesc = open( mDeviceName.c_str(), O_RDONLY|O_NONBLOCK );
	if ( mFiledesc == -1 )
	{
		// Try alternate/old device file.
		mFiledesc = open( mDeviceName2.c_str(), O_RDONLY|O_NONBLOCK );
	}

	BEGIN_ONCE
	{
		if ( mFiledesc == -1 )
		{
			CERROR << "Can't open joystick device at "
                   << mDeviceName << " nor " << mDeviceName2 << '.' << endl
                   << "Try modprobe joydev." << endl;
		#if OS_FREEBSD
			CERROR << "On FreeBSD, install the port devel/linux-js" << endl
			       << "which emulates the Linux joystick driver /dev/input/js0" << endl;
		#endif
		}
	} END_ONCE

	mEnabled = (mFiledesc != -1);
	return mEnabled;
}

/*****************************************************************************
 * Close Linux joystick device.
 *****************************************************************************/
void
JoystickLinux::Close( void )
{
	if ( mFiledesc != -1 )
		close( mFiledesc );
	mFiledesc = -1;  // remember it's closed
	mEnabled = false;
}

/*****************************************************************************
 * Dequeue one event from Linux driver and enqueue it into InputQueue.
 * @return False if no more events.
 *****************************************************************************/
bool
JoystickLinux::PollEvent( Event& event /*OUT*/ )
{
	// Base Joystick ensures this isn't recursed (busy flag).

	event.Clear();

	// Button events have priority over axis events.
	if ( not IfEnabled() )
		return false;

	// Read the next event from the Linux joystick driver.
	struct js_event eventLinux;
	int bytesRead = read( mFiledesc, &eventLinux, sizeof(struct js_event) );
	if ( bytesRead != sizeof(struct js_event) )
	{
		return false;
	}

	// Button event?
	if ( eventLinux.type == JS_EVENT_BUTTON )
	{
		event.mButton    = eventLinux.number;   // zero based
		event.mButtonVal = eventLinux.value;  // bool: pressed/released
		return true;
	}

	// An axis event?
	if ( eventLinux.type == JS_EVENT_AXIS )
	{
		event.mAxis    = eventLinux.number;  // zero based
		event.mAxisVal = eventLinux.value;
		return true;
	}

	return false;
}

/*****************************************************************************
 * Print status of joystick.
 *****************************************************************************/
void
JoystickLinux::PrintStatus( void )
{
	if ( not BASE_CONF.mQuiet && IfEnabled() )
	{
#if OS_FREEBSD
		CLOG << "Enabled Linux joystick (on FreeBSD)." << endl;
#else
		CLOG << "Enabled Linux joystick." << endl;
#endif
	}
}

} // namespace input

#endif // OS_LINUX || OS_FREEBSD
