/*****************************************************************************
 * $LastChangedDate: 2008-12-12 21:44:59 -0500 (Fri, 12 Dec 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Interface between C++ and Lua scripts.
 * @remarks module.txt documents how to export new C++ classes to Lua.
 *//*
 * LEGAL:   COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef LUABIND_LUA_HH
#define LUABIND_LUA_HH 1

#define GET_LUA() Lua::GetInstance()

#include "lua_bind/lua_lang.hh"
#include "lua_lang/luaconf.h"

namespace lua {

/// @brief Exception in Lua.
class LuaException : public Exception
{
public:
	LuaException( void ) : Exception() { }
	LuaException( string text ) : Exception(text) { }
};

////////////////////////////////////////////////////////////////////////////////
/// @brief Lua interpreter.
///
/// NOTE: Most methods might throw a C++ exception.
///
class Lua
{
PREVENT_COPYING(Lua)
private:
	Lua( void );
	~Lua();

public:
    CLASS_METHOD Lua& GetInstance( void )  // Singleton
	{
		if ( UX( msInstance == NULL ) )
			msInstance = new Lua;
		return *msInstance;
	}

	typedef LUA_NUMBER Number;  ///< Lua numeric type

	void		ExportCmdLine( int argc, char** argv );
	void		SetScriptDir( const string& scriptDir );  // throw
	string		GetScriptDir( void ) { return mScriptDir; }
	void		RunScript( const string& scriptName, const bool recompile = false );  // throw
	bool		ReadVarAsBool( const string& varName );  // throw
	Number		ReadVarAsNumber( const string& varName );  // throw
	string		ReadVarAsString( const string& varName );  // throw
	Number		ReadTableAsNumber( const string& tableName, const int tableIdx );  // throw
	Number		ReadTableAsNumber( const string& tableName, const string& memberName );  // throw
	string		ReadTableAsString( const string& tableName, const int tableIdx );  // throw
	string		ReadTableAsString( const string& tableName, const string& memberName );  // throw
    Number      CallLuaFunc( const string& tableName, const string& funcName, const vector<Lua::Number>& args );

private:
	void SaveStack( void )
	{
		mSavedStack = lua_gettop( mLua );
	}

	void RestoreStack( void )
	{
		lua_settop( mLua, mSavedStack );
	}

	// Throws a C++ exception but first restores Lua's stack.
	void ThrowLuaException( const string& msg )
	{
		RestoreStack();
		throw LuaException( msg );
	}

private:
	// mFuncTable[] is defined at lua_bind/funcs.cc
	typedef map<string,int>		BytecodeMap;
    CLASS_VAR Lua*				msInstance;		///< instance of Lua
	CLASS_VAR const luaL_Reg	mFuncTable[];	///< C interface functions exported to Lua scripts
	lua_State*					mLua;			///< the Lua interpreter (C code)
	string						mScriptDir;		///< directory containing scripts
	BytecodeMap					mBytecodeMap;	///< avoids recompiling Lua bytecode
	int							mSavedStack;	///< stack depth
};

} // namespace lua

#endif // LUABIND_LUA_HH
