/*****************************************************************************
 * $LastChangedDate: 2008-12-06 15:09:15 -0500 (Sat, 06 Dec 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Cloud particle system.
 *//*
 * LEGAL:   COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define SKY_CLOUDS_CC 1
#include "base/module.hh"
#include "base/random.hh"
using namespace base;
#include "math/module.hh"
#include "math/graphics.hh"
using namespace math;
#include "glue/module.hh"
#include "glue/sim_time.hh"
using namespace glue;
#include "view/module.hh"
using namespace view;
#include "fx/module.hh"
#include "fx/texture.hh"
using namespace fx;
#include "world/module.hh"
using namespace world;
#include "sky/module.hh"
#include "sky/clouds.hh"
#include "sky/conf.hh"
#include "sky/sky_dome.hh"

namespace sky {

// Area of sky populated with clouds is reduced to avoid overlapping sky dome.
const fp SKY_WIDTH_RATIO = 0.75f;

const StateBits CLOUD_STATE_BITS = STATE_BIT_BLEND
                                 | STATE_BIT_TRANSPARENT
                               //| STATE_BIT_NO_DEPTH_TEST
                                 | STATE_BIT_NO_LIGHTING
                                 | STATE_BIT_NO_FOG;

////////////////////////////////////////////////////////////////////////////////
/// @brief Listener to rotates clouds for every view.
///
CLASS_EVENT_LISTENER( CloudPreDrawViewListener, shptr<View>, CloudParticleSystem, mCloudParticleSystem )

////////////////////////////////////////////////////////////////////////////////
////////////////////////////  CloudParticle  ///////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
CloudParticle::CloudParticle( const Sprite::Args& spriteArgs )
:	BigParticle(spriteArgs,BigParticle::UNLIMITED_LIFETIME)
{
	// NOP
}

CloudParticle::~CloudParticle()
{
	// NOP
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////  Clouds (plural)  /////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 * For speed, timer-tick is disabled (so clouds won't move).
 *****************************************************************************/
CloudParticleSystem::CloudParticleSystem( SkyDome& skyDome, const BoxVolume& skyVolume )
:	BigParticleSystem(BigParticleSystem::Args(BigParticle::UNLIMITED_LIFETIME,
                                              CLOUD_STATE_BITS,
                                              TIMER_DISABLED)),  // default stateBits won't be used, since  
	mPreDrawViewListener(new CloudPreDrawViewListener(*this)),		       // passed for specific cloud textures later
	mRollDeg(0.0f)
{
CHECK_TYPESIG(&skyDome,TYPESIG_SKY_DOME);

	// Clouds are too white-and-bright for night.
	if ( SKY_CONF.mCloudsEnabled
         and not GET_SIM_TIME().IfNightTime() )
	{
		MakeClouds( skyDome, skyVolume );
	}

	// Listener to rotate cloud sprites before rendering a view.
	// Register functors to be called before rendering this View.
	EVENT_VIEW_PRE_DRAW_ONE_VIEW.Listen( mPreDrawViewListener.PTR() );
}

CloudParticleSystem::~CloudParticleSystem()
{
	// Unregister or crash into destroyed object!
	EVENT_VIEW_PRE_DRAW_ONE_VIEW.Unlisten( mPreDrawViewListener.PTR() );
}

/*****************************************************************************
 * ctor subroutine.
 *****************************************************************************/
void
CloudParticleSystem::MakeClouds( SkyDome& skyDome, const BoxVolume& skyVolume )
{
CHECK_TYPESIG(&skyDome,TYPESIG_SKY_DOME);

	// Load cloud textures
	// Create their own graphical states for state-sorting.
	vector<StateBits>              stateBits;
	vector<RefPtr<osg::StateSet> > stateSets;

#define MAKE_CLOUD_STATE( TEXTURE_NAME )											\
{																					\
	stateBits.push_back( CLOUD_STATE_BITS | StateSorter::AllocateStateBitsId() );	\
	stateSets.push_back( Texture::NewTextureAsStateSet( TEXTURE_NAME ) );			\
}

	MAKE_CLOUD_STATE( "textures/cloud0.png" );
	MAKE_CLOUD_STATE( "textures/cloud1.png" );
	MAKE_CLOUD_STATE( "textures/cloud2.png" );
	MAKE_CLOUD_STATE( "textures/cloud3.png" );

ASSERT( stateBits.size() == stateSets.size() );

	const bool flatWorld   = GET_WORLD().IfFlat();
	sky::Conf::CloudConf& cloudConf = flatWorld ? SKY_CONF.mCloudConfFlat
                                                : SKY_CONF.mCloudConfSphere;
	const uint cloudCnt    = cloudConf.mCloudCnt;
	const fp skyWidth      = ABS( skyVolume.mMax[LAT] - skyVolume.mMin[LAT] ) * SKY_WIDTH_RATIO;
	const fp skyWidthHalf  = skyWidth * 0.5f;
	const fp altMin        = flatWorld ? Geo::Meters2Sim( cloudConf.mCloudMinAlt ) + skyVolume.mMin[ALT]
                                       : Geo::Meters2Sim( cloudConf.mCloudMinAlt );
	const fp altMax        = flatWorld ? Geo::Meters2Sim( cloudConf.mCloudMaxAlt ) + skyVolume.mMax[ALT]
                                       : Geo::Meters2Sim( cloudConf.mCloudMaxAlt );
	const fp altDelta      = altMax - altMin;
	const fp widthMin      = Geo::Meters2Sim( cloudConf.mCloudMinWidth );
	const fp widthDelta    = Geo::Meters2Sim( cloudConf.mCloudMaxWidth ) - widthMin;
	WorldVertex pos;

ASSERT( cloudCnt < 50*THOUSAND );
ASSERT( skyWidth > 0 );
ASSERT( altDelta > 0 );
ASSERT( widthDelta > 0 );

    // Loop to create cloud sprites.
    // Since some cloud args will be rejected (outside skydome),
    // the number of loops will be greater than cloudCnt.
    uint curCloudCnt, i;  // curCloudCnt counts each actual cloud
	for ( curCloudCnt = i = 0;
          (curCloudCnt < cloudCnt) and (i < cloudCnt * 4);
          ++i )
	{
		if ( flatWorld )
		{
			pos[XX] = gRandom.random_f(skyWidth) - skyWidthHalf;
			pos[YY] = gRandom.random_f(skyWidth) - skyWidthHalf;
            pos[ZZ] = altMin + gRandom.random_f(altDelta);
		}
		else
		{
			SphereVertex sv( gRandom.random_f(180.0f) - 90.0f,
			                 gRandom.random_f(360.0f) - 180.0f,
                             Meter(altMin + gRandom.random_f(altDelta)) );
			pos = Geo::SphereVertex2WorldVertex( sv );
		}

        // Skip if cloud would be outside sky-dome.
        const fp width = widthMin + gRandom.random_f( widthDelta );
        if ( skyDome.IfInside( pos, width * 0.5f ) )
        {
            // Create a cloud sprite.
            ++curCloudCnt;

    		// Randomly pick a cloud texture.
        	uint idxCloud = 0;  // favor the regular cloud texture
        	if ( gRandom.random_ui(100) > 50 )
            	idxCloud = gRandom.random_ui( stateSets.size() );
            ASSERT( idxCloud < stateSets.size() );
            ASSERT( idxCloud < stateBits.size() );

    		const Sprite::Args spriteArgs( pos,
                                           width,
                                           stateSets[idxCloud],
                                           RGBA(0xff,0xff,0xff,0x80) );

    		AttachParticle( new CloudParticle( spriteArgs ),
                            stateBits[idxCloud],
                            stateSets[idxCloud] );
        }
	}
}

/*****************************************************************************
 * Rotate clouds according to view.
 * As an illusion, clouds sprites are counter-rolled as player rolls.
 *****************************************************************************/
void
CloudPreDrawViewListener::operator()( shptr<View> view )
{
	// Reflect.
	mCloudParticleSystem.PreDraw( view );
}

void
CloudParticleSystem::PreDraw( shptr<View> view )
{
	// Every View tracks its "roll degree" when its view matrix is recomputed.
	// Rotating all the Sprites is very slow, so bypass if change in roll degree is nearly zero.
	const Degree rollDeg = view->GetRollDegree();
	if ( ABS( rollDeg - mRollDeg ) > 0.1f )
	{
		mRollDeg = rollDeg;  // current degree of roll of all cloud sprites
		const Radian rollRad = Deg2Rad( rollDeg );

		// Rotate (or not if the roll angle is zero) all the cloud sprites.
		for ( BigParticles::iterator iter = mBigParticles.begin();
	          iter != mBigParticles.end();
	          ++iter )
		{
			// As an illusion, clouds sprites are counter-rolled as player rolls.
			(*iter)->RotateRoll( rollRad, Sprite::ROTATE_ABS );  // Particle->RotateRoll()
		}
	}
}

} // namespace sky
