/*****************************************************************************
 * $LastChangedDate: 2008-12-24 10:42:02 -0500 (Wed, 24 Dec 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Sky model (Singleton).
 *//*
 * LEGAL:   COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define SKY_SKY_CC 1
#include "base/module.hh"
using namespace base;
#include "gfx/module.hh"
#include "gfx/osg.hh"
#include "gfx/light_nums.hh"
using namespace gfx;
#include "glue/module.hh"
#include "glue/sim_time.hh"
using namespace glue;
#include "graph/module.hh"
#include "graph/light.hh"
using namespace graph;
#include "view/module.hh"
#include "view/window.hh"
using namespace view;
#include "world/module.hh"
using namespace world;
#include "sky/module.hh"
#include "sky/sky.hh"
#include "sky/conf.hh"
#include "sky/defs.hh"
#include "sky/sky_dome.hh"
#include "sky/sky_dome_painter.hh"
#include "sky/clouds.hh"
#include "sky/sun.hh"
#include "sky/moon.hh"
#include "sky/stars.hh"

namespace sky {

const RGBA SKY_LIGHT_BASE_AMBIENT(     0.05f, 0.05f, 0.05f, 1.00f );
const RGBA DEFAULT_LIGHT_BASE_AMBIENT( 0.05f, 0.05f, 0.05f, 1.00f );
const RGBA FOG_COLOR_NIGHT(            0.20f, 0.20f, 0.45f, 1.00f );

////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////  Sky  ///////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

Sky* Sky::msInstance;

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Sky::Sky( void )
:	mSkyDome(new SkyDome( gWorld.GetPosition(),
                          gWorld.GetRadius(),
                          SKY_CONF.mSkyColor,
                          SKY_CONF.mSkyDomeSliceCnt ) ),  // -- ORDER-DEPENDENT --
	mCloudParticleSystem(new CloudParticleSystem(*mSkyDome,GetVolume())),
	mSun(new Sun( sky::defs::DEFAULT_SUN_LAT, sky::defs::DEFAULT_SUN_LON ) ),
	mMoon(new Moon( *mSkyDome,
                    sky::defs::DEFAULT_MOON_LAT,
                    sky::defs::DEFAULT_MOON_LON ) ),
	mStars(new Stars( gWorld.GetPosition(),
                      SKY_CONF.mStarCnt,
                      *mSkyDome ) ),
	mLight(new Light( gfx::defs::LIGHT_NUM_SKY,
		              WorldVertex(0,0,0),  // Lua will set position appropriately
				      RGBA( 0.9f, 0.5f, 0.5f, 1.0f ),     // ambient
				      RGBA( 0.5f, 0.5f, 0.5f, 1.0f ),     // diffuse
				      RGBA( 0.5f, 0.5f, 0.5f, 1.0f ) ) )  // specular
{
CHECK_TYPESIG(mSkyDome,TYPESIG_SKY_DOME);
ASSERT( gWorld.GetRadius() > 0.0f );
ASSERT( SKY_CONF.mSkyDomeSliceCnt > 0 );

	// NOP.  The ctor creates the Objects.
	// Later, Sky::Start() will attach them to the scene graph.

#if DEBUG
{
	// Self-check.
	pair<Degree,Degree> latLon;
	latLon = ComputeLatLonFromHour(  0.0f ); ASSERT( FP_EQ<fp>( latLon.first, 180.0f ) ); ASSERT( FP_EQ<fp>( latLon.second, 0.0f ) );
	latLon = ComputeLatLonFromHour(  6.0f ); ASSERT( FP_EQ<fp>( latLon.first,  90.0f ) ); ASSERT( FP_EQ<fp>( latLon.second, 0.0f ) );
	latLon = ComputeLatLonFromHour( 12.0f ); ASSERT( FP_EQ<fp>( latLon.first,   0.0f ) ); // lon doesn't matter at poles
	latLon = ComputeLatLonFromHour( 18.0f ); ASSERT( FP_EQ<fp>( latLon.first,  90.0f ) ); ASSERT( FP_EQ<fp>( latLon.second, 180.0f ) );
}
#endif
}

Sky::~Sky()
{
	// NOP
}

/*****************************************************************************
 * Attach the sky nodes to the scene graph.
 *****************************************************************************/
void
Sky::Start( const fp hour )
{
CHECK_TYPESIG(mSkyDome,TYPESIG_SKY_DOME);

BEGIN_ONCE {

ASSERT_HOUR(hour);

	// Apply time-of-day to components of sky.
	SetHour( hour );

	// Attach the Object components of the sky to the scene graph
	// in appropriate state-sorted group nodes.
	// TODO/FIXME: Sky components aren't suitable for OSSIM yet.
	if ( GET_WORLD().IfFlat() )
	{
		GET_WORLD().Attach( mSkyDome.PTR(), mSkyDome->GetStateBits() );
		GET_WORLD().Attach( mMoon.PTR(),    mMoon->GetStateBits() );
		GET_WORLD().Attach( mStars.PTR(),   mStars->GetStateBits() );
	}

	// Initialize lighting and fog.
	InitLightingAndFog( hour );

} END_ONCE
}

/*****************************************************************************
 * Initialize lighting and fog.
 *****************************************************************************/
void
Sky::InitLightingAndFog( const fp hour )
{
BEGIN_ONCE {

	// Stateset of the root node.
	RefPtr<osg::StateSet> stateSet = GET_SCENE_GRAPH_ROOT_STATESET();

	// Enable or disable lighting.
	if ( SKY_CONF.mLightEnabled )
	{
		// Attach light source to scene graph.
		// TODO/FIXME: Position of light should be same as the position of sun.
		GET_SCENE_GRAPH()->InsertGroupNode( mLight->GetGroupNode() );

		// Set OpenGL's global ambient lighting.
		mLight->SetBaseAmbient(           SKY_LIGHT_BASE_AMBIENT );
		Light::SetDefaultBaseAmbient( DEFAULT_LIGHT_BASE_AMBIENT );
	}
	else
	{
		// Disables lighting all thru the graph.
		stateSet->setMode( GL_LIGHTING, osg::StateAttribute::OFF | osg::StateAttribute::OVERRIDE );
	}

	// Enable fog.
	if ( SKY_CONF.mFogEnabled )
	{
		RefPtr<osg::Fog> mFog = new osg::Fog;
        if ( SimTime::IfNightTime( hour ) )
    		mFog->setColor( FOG_COLOR_NIGHT );
        else
    		mFog->setColor( RGBA2VEC4F( SKY_CONF.mFogColor ) );
		mFog->setDensity( SKY_CONF.mFogDensity );
		mFog->setMode( SKY_CONF.mFogMode );
		stateSet->setAttributeAndModes( mFog.get() );
		stateSet->setMode( GL_FOG, osg::StateAttribute::ON );
	}

} END_ONCE
}

/*****************************************************************************
 * Update sky for time-of-day.
 *****************************************************************************/
void
Sky::SetHour( const fp hour )
{
ASSERT_HOUR(hour);

	// Changes Light objects.
	Light::SetMulColor( SimTime::IfNightTime(hour) ? 0.3f : 1.0f );

	// Change position of sun.
#if 0
	mSun->SetLatLon( ComputeLatLonFromHour( hour ) );
#else
    // Make sun visible thru particular random arrangement of clouds.
	mSun->SetLatLon( make_pair<>( Degree(82.0f), Degree(40.0f) ) );
#endif

	// Re-paint the sky dome.
	SkyDomePainter::Paint( *mSkyDome, hour );

    // For OSSIM, force dark-blue sky always.
	if ( not GET_WORLD().IfFlat() )
    {
        GET_SCENE_GRAPH()->SetClearColor( RGBA( 0.02f, 0.02f, 0.30f, 1.0f ) );
    }
}

/*****************************************************************************
 * @return Sky's Light.
 *****************************************************************************/
shptr<Light>
Sky::GetLight( void )
{
CHECK_TYPESIG(mLight,TYPESIG_LIGHT);

	return mLight;
}

/*****************************************************************************
 * Compute (lat,lon) of sun from time-of-day.
 *****************************************************************************/
pair<Degree,Degree>
Sky::ComputeLatLonFromHour( const fp hour )
{
ASSERT_HOUR(hour);

	// hour  0: (lat=180,lon=0)
	// hour  6: (lat= 90,lon=0)
	// hour 12: (lat=  0,lon=0 or 180)
	// hour 18: (lat= 90,lon=180)
	// hour 24: (lat=180,lon=0 or 180)
	const Degree lat = ABS(hour - 12.0f) * (180.0f / 12.0f);
	const Degree lon = (hour < 12.0f) ? 0.0f : 180.0f;
	ASSERT_LAT_LON_DEG( lat, lon );
	return make_pair<Degree,Degree>( lat, lon );
}

/*****************************************************************************
 * @return Volume of sky in 3D space.
 *****************************************************************************/
BoxVolume
Sky::GetVolume( void )
{
	return GET_WORLD().GetVolume();
}

/*****************************************************************************
 * Accessors.
 *****************************************************************************/
SkyDome&
Sky::GetSkyDome( void )
{
CHECK_TYPESIG(mSkyDome,TYPESIG_SKY_DOME);

	return *mSkyDome;
}

Sun&
Sky::GetSun( void )
{
	return *mSun;
}

} // namespace sky
