/*****************************************************************************
 * $LastChangedDate: 2008-12-26 21:33:34 -0500 (Fri, 26 Dec 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Main sound header file and abstract Sound class.
 * @namespace sound
 * @brief Sound.
 *//*
 * LEGAL:	COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef SOUND_SOUND_HH
#define SOUND_SOUND_HH 1

namespace sound {

const fp DEFAULT_VOLUME = 1.0;

////////////////////////////////////////////////////////////////////////////////
/// @brief Abstract Sound class (also used as a NOP class).
///
/// Sounds are identified/keyed by filename.
///
/// Sound class:
/// ------------
/// Sound class is a Facade over a sound libary.
/// If sound library isn't available, Sound class will be compiled as NOPs.
///
/// Disabling Sound:
/// ----------------
/// Sound can only be disabled at program startup (before Sound::GetInstance()).
/// Once disabled, it stays disabled.
///
/// Changing volume of a sample:
/// ----------------------------
/// Changing the volume of a sample should be considered a slow operation.
///
/// Anti-redundancy and tracking of playing:
/// ----------------------------------------
/// Play() will become a NOP if redundantly passed the same args while it is still playing.
/// So clients can rely on this Sound class to keep track of which samples are playing.
///
class SoundAbstract
{
public:
	enum { PLAY_ONCE = false,
           PLAY_LOOP = true };

protected:
				SoundAbstract( void ) { }
	virtual		~SoundAbstract( void ) { }

public:
	CLASS_METHOD SoundAbstract&	GetInstance( void )
	{ PERSISTENT SoundAbstract sDummy; return sDummy; }  // but it works
	CLASS_METHOD void			Disable( void ) { }  // don't use with GetInstance()
	virtual bool				IfPlaying( const string& filename ) { return false; }
	virtual bool				IfPlaying( const string& filename, const fp volume ) { return false; }
	// Play() is a NOP if sample is already playing with the same settings.
	virtual bool				Play( const string& filename, const bool loop, const fp volume = DEFAULT_VOLUME )
                                { return true; }
	virtual void				Stop( const string& filename ) { }
	virtual void				StopAll( void ) { }
	virtual void				SetVolume( const string& filename, const fp volume ) { }
};

} // namespace sound

//------------------------------------------------------------------------------

#if SOUND_PLIB
#include "sound/sound_plib.hh"
#endif

//------------------------------------------------------------------------------

namespace sound {

#if SOUND_PLIB
	typedef SoundPlib Sound;
#else
	typedef SoundAbstract Sound;  // abstract class also used as a NOP class
#endif

} // namespace sound

#endif // SOUND_SOUND_HH
