/*****************************************************************************
 * $LastChangedDate: 2008-10-06 10:35:00 -0400 (Mon, 06 Oct 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Sound class based on PLIB.
 * @verbatim
 *
 * PLIB sound:
 * -----------
 * - To play sound, PLIB requires being pulsed (5 Hz minimum?).
 *   Otherwise, it will play a snippet of an audio file, then fade out.
 * - PLIB alters a sample's volume by modifying its waveform (slow operation).
 *   This Sound class avoids PLIB distorting the waveform by restoring
 *   the original waveform (so changing volume becomes even slower).
 * - PLIB will cause program exit if a sample is deleted while still playing.
 *   stopSample() and update() don't quite work according to the PLIB docs.
 *   The reliable way is to enqueue zombie samples then deleted them
 *   when their "play count" reaches zero.
 *
 * @endverbatim
 *//*
 * LEGAL:	COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#if SOUND_PLIB

#if DEBUG
#ifndef SOUND_DEBUG
#define SOUND_DEBUG 0  // debug
#endif // SOUND_DEBUG
#endif // DEBUG

#define SOUND_SOUND_PLIB_CC 1
#include "base/module.hh"
#include "base/file.hh"
#include "base/timer.hh"
#include "base/conf.hh"
using namespace base;
#include "sound/module.hh"
#include "sound/sound_plib.hh"
// after base
#include <plib/sl.h>
#include <plib/sm.h>

namespace sound {

INTERN const uint   SOUND_SAMPLE_RATE	 = 8000;
#define             SOUND_MIXER_DEVICE   "/dev/mixer"
INTERN const uint   SOUND_PLIB_VOLUME	 = 50;
INTERN Milliseconds SOUND_PLIB_TICK_FREQ = 20;

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////  SoundPlib  ///////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

SoundPlib* SoundPlib::msInstance;
bool       SoundPlib::msEnabled = true;  // if disabled, it stays disabled, Enable() omitted

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
SoundPlib::SoundPlib( void )
{
	if ( msEnabled )
	{
		mSched = new slScheduler( SOUND_SAMPLE_RATE );
		mMixer = new smMixer( SOUND_MIXER_DEVICE );
		mMixer->setMasterVolume( SOUND_PLIB_VOLUME );

		// PLIB requires a pulse
		Timer::GetInstance().Register( new TickFunctor, SOUND_PLIB_TICK_FREQ );
	}
	else
	{
		mSched = NULL;  // this is a reason why Sound cannot be enabled if initially disabled
		mMixer = NULL;
	}
}

SoundPlib::~SoundPlib()
{
	DELETE_NULL( mSched );
	DELETE_NULL( mMixer );
}

/*****************************************************************************
 * NOTE: Disable() is a class method.
 *       DO NOT CALL WITH GetInstance().
 * Enable() is omitted.
 * If Disable() is ever called, it should be done during startup,
 * before SoundPlib::GetInstance() is called.
 *****************************************************************************/
void  // CLASS_METHOD
SoundPlib::Disable( void )
{
// Shouldn't be instantiated yet since the ctor enables sound.
// Disable() should be called as a class method (without GetInstance()).
ASSERT( msInstance == NULL );

	msEnabled = false;
}

/*****************************************************************************
 * @return True if an audio sample is playing.
 * Ignores volume setting.
 *****************************************************************************/
bool
SoundPlib::IfPlaying( const string& filename )
{
	return IfPlaying( filename, -1.0f );
}

/*****************************************************************************
 * @return True if an audio sample is playing.
 * @param	filename
 * @param	volume
 *			Volume ignored if < 0 (only compare filename of sample).
 *****************************************************************************/
bool
SoundPlib::IfPlaying( const string& filename, const fp volume )
{
if ( UX( not msEnabled ) )
	return false;

	// Samples are cached in a STL map.
	// They may be still playing or finished playing.
	SampleMap::iterator iter = mSampleMap.find( filename );
	if ( iter != mSampleMap.end() )
	{
		// Was played.  Still playing?
		CHECK_TYPESIG(iter->second,TYPESIG_SOUND_SAMPLE);
		if ( iter->second->IfPlaying() )  // second=shptr<Sample>
		{
			// Ignore volume if < 0.
			if ( volume >= 0.0f )
				return volume == iter->second->GetVolume();
			else
				return true;
		}
		else
		{
			// Finished playing.
			return false;
		}
	}
	else
	{
		// Never loaded so it isn't playing.
		return false;
	}
}

/*****************************************************************************
 * Play audio file.
 * @param	filename
 *			Identifies sound file.
 * @param	loop
 *			Play continuously.
 * @param	volume
 *			Default is DEFAULT_VOLUME.
 * @return false if error (else true, even if sound is disabled).
 *****************************************************************************/
bool
SoundPlib::Play( const string& filename, const bool loop, const fp volume )
{
if ( UX( not msEnabled ) )
	return true;  // return false only if error

// NOP if the sample is already playing with the same volume.
if ( UX( IfPlaying( filename, volume ) ) )
{
#if SOUND_DEBUG
	CLOG << "SoundPlib: sample " << filename << " is playing at volume " << volume << endl;
#endif
	return true;
}

	try
	{
		// Was audio file read before?
		SampleMap::iterator iter = mSampleMap.find( filename );
		if ( iter != mSampleMap.end() )
		{
			// Play again.
			CHECK_TYPESIG(iter->second,TYPESIG_SOUND_SAMPLE);
			iter->second->Play( loop, volume );  // second=shptr<Sample>
		}
		else
		{
			// Read audio file.  Store it in STL map to speed replay.
			shptr<Sample> sample = new Sample( filename );
			mSampleMap.insert( make_pair<string,shptr<Sample> >( filename, sample ) );

			// Play.
			sample->Play( loop, volume );
		}
	}
	catch ( Exception& e )  // Sample ctor might throw
	{
		CERROR << "ERROR: cannot play sound file " << filename << endl;
		return false;
	}

	return true;
}

/*****************************************************************************
 * Stop playing sample.
 *****************************************************************************/
void
SoundPlib::Stop( const string& filename )
{
if ( UX( not msEnabled ) )
	return;

	SampleMap::iterator iter = mSampleMap.find( filename );
	if ( iter != mSampleMap.end() )
	{
		CHECK_TYPESIG(iter->second,TYPESIG_SOUND_SAMPLE);
		iter->second->Stop();  // Sample::Stop()
	}
}

/*****************************************************************************
 * Stop playing all samples.
 *****************************************************************************/
void
SoundPlib::StopAll( void )
{
if ( UX( not msEnabled ) )
	return;

	for ( SampleMap::iterator iter = mSampleMap.begin();
          iter != mSampleMap.end();
          ++iter )
	{
		CHECK_TYPESIG(iter->second,TYPESIG_SOUND_SAMPLE);
		iter->second->Stop();  // Sample::Stop()
	}
}

/*****************************************************************************
 * Set volume of sample.
 * @param	volume
 *			0.0 = off, 1.0 max volume.
 *****************************************************************************/
void
SoundPlib::SetVolume( const string& filename, const fp volume )
{
if ( UX( not msEnabled ) )
	return;

	SampleMap::iterator iter = mSampleMap.find( filename );
	if ( iter != mSampleMap.end() )
	{
		CHECK_TYPESIG(iter->second,TYPESIG_SOUND_SAMPLE);
		iter->second->SetVolume( volume );  // Sample::SetVolume()
	}
}

/*****************************************************************************
 * Timer-tick pulses the plib sound system.
 *****************************************************************************/
void
SoundPlib::TickFunctor::operator()( const Milliseconds millisecElapsed )
{
	SoundPlib::GetInstance().Pulse();
}

void
SoundPlib::Pulse( void )
{
if ( UX( not msEnabled ) )
	return;

	// Call PLIB to play next segment of waveform.
	mSched->update();

	// Destroy zombie samples after they're finished playing.
	// PLIB will exit program if sample were to be deleted while still playing.
	for ( set<slSample*>::iterator iter = mZombieSamples.begin();
          iter != mZombieSamples.end();
          /*NOP*/ )
	{
		slSample* sample = *iter;
		if ( sample->getPlayCount() == 0 )  // if done playing
		{
			// Delete PLIB sample.
			delete sample;

			// STL trick to delete node.
			set<slSample*>::iterator iter2 = iter;
			++iter2;
			mZombieSamples.erase( iter );
			iter = iter2;
		}
		else
		{
			++iter;
		}
	}
}

/*****************************************************************************
 * Enqueue a zombie sample (possibly still playing).
 *****************************************************************************/
void
SoundPlib::EnqueueZombieSample( slSample* zombie )
{
	mZombieSamples.insert( zombie );
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////  SoundPlib::Sample  ///////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Load sample from file.
 * @param	filename
 *			gzipped audio file (eg crash.wav.gz).
 *****************************************************************************/
SoundPlib::Sample::Sample( const string& filename, const fp volume )
:	mVolume(volume),
	mLoop(false),
	mSLSample(NULL),
	mBuf()
{
	SET_TYPESIG(this,TYPESIG_SOUND_SAMPLE);

	// Catenate pathname of sound file.
	string pathname = string(GetDataDir()) + string("/sounds/") + string(filename);

	// Read gzipped sound file into C++ string mBuf.
	if ( ReadFileGzip( pathname, mBuf ) )
	{
	ASSERT( mBuf.size() > 0 );

		// Create PLIB sound object.
		// PLIB will copy the buffer (I read the source, Luke).
		mSLSample = new slSample( reinterpret_cast<const uchar*>(mBuf.c_str()), mBuf.size() );

		// Change volume unless default.
		if ( volume != DEFAULT_VOLUME )
		{
			mSLSample->adjustVolume( volume );
		}
	}
	else
	{
		mSLSample = NULL;
		throw Exception( string("ERROR: cannot read sound file ") + string(filename) );
	}

ASSERT( mSLSample != NULL );
ASSERT( mBuf.size() > 0 );
}

SoundPlib::Sample::~Sample()
{
	DELETE_NULL( mSLSample );
	INVALIDATE_TYPESIG(this,TYPESIG_SOUND_SAMPLE);
}

/*****************************************************************************
 * Play sample.
 *****************************************************************************/
void
SoundPlib::Sample::Play( const bool loop, const fp volume )
{
CHECK_TYPESIG(this,TYPESIG_SOUND_SAMPLE);

	// Remember.
	mLoop = loop;

	// Adjust volume before playing to avoid PLIB
	// modifying waveform for volume adjustment while playing.
	SetVolume( volume );
ASSERT( mVolume == volume );

	// Play one-shot or loop.
	if ( loop )
		SoundPlib::GetInstance().mSched->loopSample( mSLSample );
	else
		SoundPlib::GetInstance().mSched->playSample( mSLSample );
}

/*****************************************************************************
 * Stop playing sample.
 *****************************************************************************/
void
SoundPlib::Sample::Stop( void )
{
CHECK_TYPESIG(this,TYPESIG_SOUND_SAMPLE);

	SoundPlib::GetInstance().mSched->stopSample( mSLSample );
}

/*****************************************************************************
 * If sample is playing.
 *****************************************************************************/
bool
SoundPlib::Sample::IfPlaying( void )
{
CHECK_TYPESIG(this,TYPESIG_SOUND_SAMPLE);

	return mSLSample->getPlayCount() > 0;
}

/*****************************************************************************
 * Set volume of sample.
 * NOTE: Stops the current sample, creates a new sample that won't be playing.
 *       SoundPlib::SetVolume() is supposed to play it afterwards.
 * @param	volume
 *			0.0 = off, 1.0 max volume.
 *****************************************************************************/
void
SoundPlib::Sample::SetVolume( const fp volume )
{
CHECK_TYPESIG(this,TYPESIG_SOUND_SAMPLE);

	// NOP unless volume need changing.
	if ( volume != mVolume )
	{
		mVolume = volume;

		// Restore original sample to avoid degradation (yuck).
		// Adjusting volume modifies the waveform (yuck).
		SoundPlib::GetInstance().mSched->stopSample( mSLSample );
		SoundPlib::GetInstance().EnqueueZombieSample( mSLSample );
		mSLSample = new slSample( reinterpret_cast<const uchar*>(mBuf.c_str()), mBuf.size() );
		mSLSample->adjustVolume( volume );

		// Resuming playing?
		if ( IfPlaying() )
		{
			if ( mLoop )
				SoundPlib::GetInstance().mSched->loopSample( mSLSample );
			else  // one-shot
				SoundPlib::GetInstance().mSched->playSample( mSLSample );
		}
	}
}

fp
SoundPlib::Sample::GetVolume( void )
{
CHECK_TYPESIG(this,TYPESIG_SOUND_SAMPLE);

	return mVolume;
}

} // namespace sound

#endif // SOUND_PLIB
