/*****************************************************************************
 * $LastChangedDate: 2008-12-09 20:16:28 -0500 (Tue, 09 Dec 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Module configuration.
 *//*
 * LEGAL:	COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define VIEW_CONF_CC 1
#include "base/module.hh"
#include "base/conf.hh"
#include "base/misc.hh"
using namespace base;
#include "types/rgba.hh"
using namespace types;
#include "view/module.hh"
#include "view/conf.hh"

namespace view {

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  Conf  ///////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Conf::Conf( void )
:	mValid(false),
	// cmd-line args 
    // (none)
	// .conf file
	mWindowWidth(800),
	mWindowHeight(600),
	mWindowTitle(""),
    mViewProjectionFOV(55.0),  // fpx
    mViewProjectionAspectRatio(1.10), // fpx
    mViewNearFarRatio(0.0002),  // fpx
	mOtwViewpointOffset(0.0f),
	mWireframe(false),
	mUseTrackball(false),
	mStatsEnabled(true)
{
}

/*****************************************************************************
 * Initialize/read configuration.
 *****************************************************************************/
void
Conf::Init( int argc, char** argv, const string& confTxtFilename )
{

BEGIN_ONCE {

	mValid = true;  // assign this first as *_CONF may expand to GetConf() which asserts it

	ParseCmdlineArgs( argc, argv );   // cmdline args parsed prior to reading .conf file
	ReadConfTxt( confTxtFilename );   // read from textual .conf file
	ParseCmdlineArgs2( argc, argv );  // cmdline args that override .conf file

} END_ONCE

}

/*****************************************************************************
 * Parse cmdline args.
 *****************************************************************************/
void
Conf::ParseCmdlineArgs( int argc, char** argv )
{
	// NOP
}

void
Conf::ParseCmdlineArgs2( int argc, char** argv )
{
	// Parse cmd-line args.
	for ( int i = 1; i < argc; ++i )
	{
		if ( STREQ( argv[i], "-wh" ) )
		{
			mWindowWidth  = Argi( i+1, argc, argv );
			mWindowHeight = Argi( i+2, argc, argv );
			CmdargDone( i, 2 );
			CLOG << "Set window width,height = " << mWindowWidth << ',' << mWindowHeight << endl;
		}
		else if ( strstr( argv[i], "-wire" )  // -wire or -wireframe
               or STREQ( argv[i], "-w" ) )
		{
			mWireframe = true;
			CLOG << "Enabled wireframe mode." << endl;
		}
		else if ( STREQ( argv[i], "-t" ) )
		{
			mUseTrackball = true;
			CLOG << "Enabled trackball mode." << endl;
		}
		else if ( STREQ( argv[i], "-no-stats" )
             /* or STREQ( argv[i], "-fast" ) */ )
		{
			// OSG statistics possibly could cause inteference.
			// -stats is omitted, rather, press F12 key.
			mStatsEnabled = false;
			CLOG << "Disabled OSG stats." << endl;
		}
	}
}

/*****************************************************************************
 * Read from .conf file.
 *****************************************************************************/
void
Conf::ReadConfTxt( const string& confTxtFilename )
{
	try
	{
		ModuleConfTxt confTxt;
		confTxt.Read( string(base::GetDataDir()) + string(confTxtFilename) );
	}
	catch (...)
	{
		CERROR << "ERROR: cannot read " << confTxtFilename << endl;
		EXIT( 1 );
	}
}

void
Conf::ModuleConfTxt::ProcessField( string& field, ifstream& conf )
{
	bool error = false;

	if ( field == "WindowWidth" )
	{
		conf >> VIEW_CONF.mWindowWidth;
	}
	else if ( field == "WindowHeight" )
	{
		conf >> VIEW_CONF.mWindowHeight;
	}
	else if ( field == "WindowTitle" )
	{
		StreamReadLine( conf, VIEW_CONF.mWindowTitle );
	}
	else if ( field == "ViewProjectionFOV" )
	{
		conf >> VIEW_CONF.mViewProjectionFOV;
	}
	else if ( field == "ViewProjectionAspectRatio" )
	{
		conf >> VIEW_CONF.mViewProjectionAspectRatio;
	}
	else if ( field == "ViewNearFarRatio" )
	{
		conf >> VIEW_CONF.mViewNearFarRatio;
	}
	else if ( field == "OtwViewpointOffset" )
	{
		conf >> VIEW_CONF.mOtwViewpointOffset;  // raw world coordinate (height)
	}

    // Errors?
	if ( error )
	{
		CERROR << "ERROR[conf]: parameter for " << field << " is invalid." << endl;
		EXIT( 1 );
	}
}

/*****************************************************************************
 * Print Conf configuration.
 *****************************************************************************/
ostream&
operator<<( ostream& strm, const Conf& conf )
{
	std::ios::fmtflags savedFlags = strm.flags();

	strm << "conf [view]:"                                                       << endl
         << "WindowWidth,Height          = " << int(conf.mWindowWidth) << ','
                                             << int(conf.mWindowHeight)       ;//<< endl
	strm.flags(savedFlags);

	return strm;
}

} // namespace view
