/*****************************************************************************
 * $LastChangedDate: 2008-09-14 15:32:26 -0400 (Sun, 14 Sep 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Window class.
 *//*
 * LEGAL:   COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef VIEW_WINDOW_HH
#define VIEW_WINDOW_HH 1

// See SceneGraph for differences in root nodes.
// SceneGraph is a low-level class.
// The higher-level World class should be used for attaching nodes of 3D models
#define GET_WINDOW()						(Window::GetInstance())
#define GET_MAIN_VIEW()						(Window::GetInstance().GetMainView())
#define GET_SCENE_GRAPH()					(Window::GetInstance().GetSceneGraph())
#define GET_SCENE_GRAPH_SUPER_ROOT_NODE()	(Window::GetInstance().GetSceneGraph()->GetSuperRootNode())
#define GET_SCENE_GRAPH_ROOT_GROUP()		(Window::GetInstance().GetSceneGraph()->GetRootGroup())
#define GET_SCENE_GRAPH_ROOT_STATESET()		(Window::GetInstance().GetSceneGraph()->GetRootStateSet())
#define POST_REDISPLAY()		// NOP

#include <osgGA/TrackballManipulator>

namespace view {

class View;

////////////////////////////////////////////////////////////////////////////////
/// @brief Window contains Views, defines the scene graph, executes main loop, broadcasts draw events.
///
class Window
{
PREVENT_COPYING(Window)
friend class TrackballKeyboardHandler;
private:
                    		Window( const WidthHeight<int> wh, const string& title );
                            ~Window();
	CLASS_METHOD Window*	CreateInstance( void );

public:
    CLASS_METHOD Window&    GetInstance( void )
	{
		// NewInstance() gets ctor args from conf.
		if ( EX( msInstance != NULL ) )
			return *msInstance;
		else
			return *(msInstance = Window::CreateInstance());

	}

// The main graphics loop:
public:
	void                    EnterMainLoop( void );

// Views:
public:
    void                    AddView( shptr<View> view, const bool mainView = false );
	shptr<View>				GetMainView( void )
							{
							ASSERT( not mViews.empty() );
								return mViews.front();
							}
private:
	void					SaveViewpoint( void );
	void					RestoreViewpoint( void );

// Scene graph:
public:
    shptr<SceneGraph>		GetSceneGraph( void ) { return mSceneGraph; }
	void					SaveSceneGraphToFile( const string& filename );

// Info:
    WidthHeight<int>        GetWindowSize( void );
	Rect<int>				GetWindowRect( void );
	uint					GetFPS( void ) { return mFPS.ComputeFPS(); }

// OSG (LIMITED USE):
public:
	enum eCamera { eCamera_3D, eCamera_2D };
	void					         OsgAddCamera( RefPtr<osg::Camera> camera, const eCamera cameraType = eCamera_3D );
	RefPtr<osgViewer::GraphicsWindow> OsgGetGraphicsWindow( void ) { return mGraphicsWindow; }
    void							 OsgSetKeyboardHandler( RefPtr<osgGA::GUIEventHandler> keyboardHandler );
	void							 OsgSetStatsKey( const int key );

private:
	void					ComputeFPS( const Milliseconds elapsedMilsec );
	void					DoOperation( void );
	void					DrawViewOutline( shptr<View> view, const RGBA outlineColor );

private:
	class Operation : public osg::Operation
	{
	public:
	    Operation( void );
		virtual ~Operation();
	    virtual void operator ()( osg::Object* );
	};

private:
	struct FPS
	{
						FPS( void );
		void			Tick( const Milliseconds elapsedMilsec );
		uint			ComputeFPS( void ) { return mFPS; }

		uint			mFPS;			///< FPS of previous second
		uint			mFrameCnt;		///< frame count of current second
		Milliseconds	mFrameStart;	///< beginning of current second
	};

private:
    typedef list<shptr<View> > Views;
    CLASS_VAR Window*						msInstance;			///< there can be only one
	shptr<SceneGraph>						mSceneGraph;		///< the amazing scene graph
	RefPtr<osgViewer::Viewer>				mViewer;			///< contains cameras
    Views									mViews;				///< Window contains one or more Views
	string									mTitle;				///< window title
	FPS										mFPS;				///< frame rate
	RefPtr<osg::GraphicsContext::Traits>	mTraits;
    RefPtr<osg::GraphicsContext>			mGraphicsContext;
    RefPtr<osgViewer::GraphicsWindow>		mGraphicsWindow;
	RefPtr<osg::Operation>					mOperation;
	RefPtr0<osgViewer::StatsHandler>		mStatsHandler;		///< statistics (might be disabled/NULL)
	RefPtr<osgGA::TrackballManipulator>		mTrackballManipulator;
};

} // namespace view

#endif // VIEW_WINDOW_HH
