/*****************************************************************************
 * $LastChangedDate: 2008-08-31 02:29:07 -0400 (Sun, 31 Aug 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Test of shptr<> template class.
 *
 * [2008/05] This test is intended to test a generic implementation
 * of an intrusive reference-count template class.
 *
 * But, optionally, if COMPILE_TEST_REF_CNT is set, then the value
 * of the reference-count will be tested also
 * (testing then becomes specific to osg::ref_ptr).
 *
 *//*
 * LEGAL:	COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define COMPILE_TEST_REF_CNT 1
#define TEST_OSG 0

#if TEST_OSG
#define SHARED_CLASS osg::Referenced
#else
#define SHARED_CLASS base::Shared
#endif

#define TESTING 1
#include "test.hh"
#include "base/module.hh"
#include "base/types.hh"
using namespace base;

#if COMPILE_TEST_REF_CNT
# if TEST_OSG
#define GET_REF_CNT( SHPTR ) ((SHPTR)->referenceCount())  // OSG
# else
#define GET_REF_CNT( SHPTR ) ((SHPTR)->GetRefCnt())
# endif
#endif

class Object : public SHARED_CLASS
{
public:
						Object( int i ) : mVal(i) { ++msInstanceCnt; }
						~Object() { --msInstanceCnt; }
	CLASS_METHOD void	Clear( void ) { msInstanceCnt = 0; }
	int					Get( void ) { return mVal; }

public:
	CLASS_VAR int	msInstanceCnt;
	int				mVal;
};

int Object::msInstanceCnt;

/*****************************************************************************
 * Test that shptr auto-destroys after leaving block.
 *****************************************************************************/
bool
Test1( void )
{
	Object::msInstanceCnt = 0;
  {
	shptr<Object> obj = new Object( 123 );
	if ( Object::msInstanceCnt != 1 ) return Fail();

#if COMPILE_TEST_REF_CNT
	if ( GET_REF_CNT(obj) != 1 ) return Fail();
#endif
  }
	if ( Object::msInstanceCnt != 0 ) return Fail();

	return true;
}

/*****************************************************************************
 * Test passing a shptr.
 *****************************************************************************/
void
Test2Sub( shptr<Object> obj )
{
	if ( Object::msInstanceCnt != 1 ) Fail();

#if COMPILE_TEST_REF_CNT
	if ( GET_REF_CNT(obj) != 2 ) Fail();
#endif
}

bool
Test2( void )
{
	Object::msInstanceCnt = 0;
  {
	shptr<Object> obj = new Object( 123 );
#if COMPILE_TEST_REF_CNT
	if ( GET_REF_CNT(obj) != 1 ) return Fail();
#endif
	if ( Object::msInstanceCnt != 1 ) return Fail();
	Test2Sub( obj );
	if ( Object::msInstanceCnt != 1 ) return Fail();
  }
	if ( Object::msInstanceCnt != 0 ) return Fail();

	return true;
}

/*****************************************************************************
 * Test storing a shptr in a persistent object.
 *****************************************************************************/
void
Test3Sub( shptr<Object> obj )
{
#if COMPILE_TEST_REF_CNT
	if ( GET_REF_CNT(obj) != 2 ) Fail();  // 2 = caller, arg
#endif

	PERSISTENT shptr<Object> sObj = obj;

#if COMPILE_TEST_REF_CNT
	if ( GET_REF_CNT(obj) != 3 ) Fail();  // 3 = caller, arg, PERSISTENT
#endif

	if ( Object::msInstanceCnt != 1 ) Fail();
}

bool
Test3( void )
{
	Object::msInstanceCnt = 0;
  {
	shptr<Object> obj = new Object( 123 );

#if COMPILE_TEST_REF_CNT
	if ( GET_REF_CNT(obj) != 1 ) Fail();  // 1 = caller
#endif

	if ( Object::msInstanceCnt != 1 ) return Fail();

	Test3Sub( obj );

#if COMPILE_TEST_REF_CNT
	if ( GET_REF_CNT(obj) != 2 ) Fail();  // 2 = caller, PERSISTENT
#endif

	if ( Object::msInstanceCnt != 1 ) return Fail();
  }
	if ( Object::msInstanceCnt != 1 ) return Fail();  // still has ref-cnt

	return true;
}

/*****************************************************************************
 * Test returning a shptr.
 *****************************************************************************/
shptr<Object>
Test4Sub( void )
{
	if ( Object::msInstanceCnt != 0 ) Fail();
	return new Object( 123 );
}

bool
Test4( void )
{
	Object::msInstanceCnt = 0;
 {
	if ( Object::msInstanceCnt != 0 ) Fail();
	shptr<Object> obj = Test4Sub();
	if ( Object::msInstanceCnt != 1 ) Fail();

#if COMPILE_TEST_REF_CNT
	if ( GET_REF_CNT(obj) != 1 ) Fail();
#endif
 }
	if ( Object::msInstanceCnt != 0 ) Fail();

	return true;
}

/*****************************************************************************
 * 
 *****************************************************************************/
int main( int argc, char** argv )
{
	TEST_MAIN_BEGIN;
	int res = 1;
	if ( Test1()
     and Test2()
     and Test3()
     and Test4()
       )
	{
		CDEBUG << endl << "Test PASSED. \n";
		res = 0;
	}
	else
	{
		CDEBUG << endl << "Test FAILED! \n";
		res = 1;
	}

	TEST_MAIN_END;
	exit( res );
}
