/*****************************************************************************
 * $LastChangedDate: 2008-08-31 02:29:07 -0400 (Sun, 31 Aug 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Test of SharedPtr template class.
 *//*
 * LEGAL:	COPYRIGHT (C) 2004 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

// Defined in test.hh for all tests.
//#define TEST_SHARED_PTR_YIELD_CPU()  TestYield()

#define TESTING 1
#include "test.hh"
#include "base/shared_ptr.hh"

#ifndef TEST_SHARED_PTR_YIELD_CPU
#error
#endif

class Int : public Shared
{
public:
						Int( int i ) : mVal(i) { ++msInstanceCnt; }
						~Int() { --msInstanceCnt; }
	CLASS_METHOD void	Clear( void ) { msInstanceCnt = 0; }
						operator int() { return mVal; }
	int					Get( void ) { return mVal; }

public:
	CLASS_VAR int	msInstanceCnt;
	int				mVal;
};
int Int::msInstanceCnt;

// Get refcnt from Shared* object.
template<typename SHARED>
int
GetRefCnt( SharedPtr<SHARED>* ptr )
{
	return ptr->mObj->mRefCnt;
}

template<typename SHARED>
int
GetRefCnt( SharedPtr<SHARED>/*fool checks.sh*/& ptr )
{
	return ptr.mObj->mRefCnt;
}

struct Counts
{
	int object;				// inc'd by constructor, dec'd by destructor
	int constructor;
	int destructor;
	int num;

	void Clear( void )
	{
		object = constructor = destructor = num = 0;
	}
};

ostream&
operator<<( ostream& strm, const Counts& obj )
{
	strm << "counts: "
         << "object = "		<< obj.object			<< " "
         << "constr = "		<< obj.constructor		<< " "
         << "destr  = "		<< obj.destructor		<< "\n";
    return strm;
}

PERSISTENT Counts counts;

const int TEST_VAL = 0xabcd;
const int MAGIC = 0x198dcb1;

////////////////////////////////////////////////////////////////////////////////
//
class Object : public Shared
{
public:
	Object( void )
	{
		++counts.object;
		++counts.constructor;
		mVal = counts.num++;
		mTestVal = TEST_VAL;
	}

	Object( int val )
	{
		++counts.object;
		++counts.constructor;
		mVal = counts.num++;
		mTestVal = val;
	}

	~Object()
	{
		--counts.object;
		++counts.destructor;
	}	

	int Magic( void ) { return MAGIC; }

	int mVal;
	int mTestVal;
};

int GetRefCnt( const Object* obj ) { return obj->mRefCnt; }
int GetRefCnt( const Object& obj ) { return obj.mRefCnt; }

// SharedPtr is intended to be passed by value (not by reference).
// refCnt : refCnt prior to call
// objCnt : objCnt prior to call
bool
FuncPassedSharedPtr( SharedPtr<Object> shared, int refCnt, int objCnt )
{
	if ( GetRefCnt(shared) != refCnt + 1 ) return Fail();  // pass arg inc'd refcnt
	if ( counts.object != objCnt ) return Fail();

	SharedPtr<Object> shared2 = shared;
	if ( shared2.PTR() != shared.PTR() ) return Fail();
	if ( GetRefCnt(shared) != refCnt + 2 ) return Fail();
	if ( counts.object != objCnt ) return Fail();
	// shared2 is auto-deleted
	return true;
}

bool
FuncPassedSharedPtrThatKeepsIt( SharedPtr<Object> shared )
{
	PERSISTENT SharedPtr<Object> sShared = shared;  // keep it
	return true;
}

SharedPtr<Object>
FuncThatReturnsSharedPtr( void )
{
	return new Object;
}

SharedPtr<Object>
FuncThatReturnsSharedPtr2( void )
{
	return SharedPtr<Object>( new Object );
}

class ClassWith1SharedPtr
{
public:
    ClassWith1SharedPtr( void )
    :   mSharedPtr1(new Object(101))
    {
    }

public:
    SharedPtr<Object> mSharedPtr1;
};

class ClassWith3SharedPtrs
{
public:
    ClassWith3SharedPtrs( void )
    :   mSharedPtr1(new Object(101)),
        mSharedPtr2(new Object(102)),
        mSharedPtr3(new Object(103))
    {
    }

public:
    SharedPtr<Object> mSharedPtr1;
    SharedPtr<Object> mSharedPtr2;
    SharedPtr<Object> mSharedPtr3;
};

//--------------------------------------------------------------------------
// Test this habit of using a PERSISTENT SharedPtr in lieue of a NULL SharedPtr.
// Ensure that ~Shared/~SharePtr don't cause problem at program exit.
class NullObject : public Shared
{
};
SharedPtr<NullObject> sNullObject = new NullObject;
//--------------------------------------------------------------------------

bool
TestSharedPtr( void )
{
 // ----------------------------
 // Test SharedPtr default constructor.
 // [2007/01] It should create an object using default ctor (it doesn't initialize as NULL).
 {
	CDEBUG << "Testing SharedPtr default constructor...\n";
	counts.Clear();

	SharedPtr<Object> nil;
	if ( counts.object != 1 ) return Fail();
	if ( counts.constructor != 1 ) return Fail();
	if ( nil.mObj == NULL ) return Fail();
	if ( nil.mObj->mTestVal != TEST_VAL ) return Fail();

    // <const T> should compile as should <T>.
    SharedPtr<const Object> const_nil;
	if ( counts.object != 2 ) return Fail();
	if ( counts.constructor != 2 ) return Fail();
	if ( const_nil.mObj == NULL ) return Fail();
	if ( const_nil.mObj->mTestVal != TEST_VAL ) return Fail();
 }
 // ----------------------------


 // ----------------------------
 // Test ctors.
 {
	CDEBUG << "Testing SharedPtr ctors...\n";
	counts.Clear();
	Object* obj = new Object;    if ( GetRefCnt(*obj) != 0 ) return Fail();
	SharedPtr<Object> sp1 = obj; if ( GetRefCnt(*obj) != 1 ) return Fail();
	SharedPtr<Object> sp2 = obj; if ( GetRefCnt(*obj) != 2 ) return Fail();
 }

 // ----------------------------
 // Test SharedPtr operator*(), PTR().
 {
	CDEBUG << "Testing operator*(), PTR()...\n";
	counts.Clear();

	const int testVal = 0x11223344;
	SharedPtr<Object> shared( new Object( testVal ) );
	if ( counts.object != 1 ) return Fail();
	if ( (*shared).mTestVal != testVal ) return Fail();

	Object* obj = shared.PTR();
	if ( obj->mTestVal != testVal ) return Fail();

	if ( shared.PTR()->mTestVal != testVal ) return Fail();
 }
 // ----------------------------


 // ----------------------------
 // Test SharedPtr operator=().
 {
	CDEBUG << "Testing operator=()...\n";
	counts.Clear();

	SharedPtr<Object> shared1( new Object( 0x1111 ) );
	SharedPtr<Object> shared2( new Object( 0x2222 ) );

	shared1 = new Object( 0x3333 );  // overwriting destroys
	shared2 = new Object( 0x4444 );

	if ( counts.object != 2 ) return Fail();
	if ( counts.destructor != 2 ) return Fail();
	if ( shared1->mTestVal != 0x3333 ) return Fail();
	if ( shared2->mTestVal != 0x4444 ) return Fail();

	// A new Shared initially has zero ref-cnt.
	// Assign it to a SharedPtr, then to another SharedPtr again.
	// Then ref-cnt should be 2.
   {
	Object* obj = new Object; if ( GetRefCnt(*obj) != 0 ) return Fail();
	SharedPtr<Object> sp1;
	SharedPtr<Object> sp2;
	sp1 = obj; if ( GetRefCnt(*obj) != 1 ) return Fail();
	sp2 = obj; if ( GetRefCnt(*obj) != 2 ) return Fail();
   }
 }
 // more operator=() test cases...
 {
	counts.Clear();

	Object* obj = new Object;
	SharedPtr<Object> shared( obj );
	SharedPtr<Object> shared2 = shared;
	SharedPtr<Object> shared3 = shared2;
	SharedPtr<Object> shared4 = shared3;

	if ( GetRefCnt(shared)  != 4 ) return Fail();  // 3 references to one shared object
	if ( GetRefCnt(shared2) != 4 ) return Fail();
	if ( GetRefCnt(shared3) != 4 ) return Fail();

	if ( shared.PTR()  != obj ) return Fail();
	if ( shared2.PTR() != obj ) return Fail();
	if ( shared2.PTR() != obj ) return Fail();
	if ( shared3.PTR() != obj ) return Fail();

	if ( shared  != obj ) return Fail();
	if ( shared2 != obj ) return Fail();
	if ( shared2 != obj ) return Fail();
	if ( shared3 != obj ) return Fail();
	if ( ! (shared  == obj) ) return Fail();
	if ( ! (shared2 == obj) ) return Fail();
	if ( ! (shared2 == obj) ) return Fail();
	if ( ! (shared3 == obj) ) return Fail();

	if ( counts.object != 1 ) return Fail();
	if ( counts.constructor != 1 ) return Fail();
	if ( counts.destructor != 0 ) return Fail();
 }
    // SharedPtr auto-vars were deleted, check.
	if ( counts.object != 0 ) return Fail();		// counts should be the same
	if ( counts.destructor != 1 ) return Fail();
 // --
 // yet more operator=() test cases...
 {
	counts.Clear();
	SharedPtr<Object> shared1 = new Object;
	shared1 = new Object;  // overwrite destroys displaced object
	shared1 = new Object;
	if ( counts.constructor != 3 ) return Fail();
	if ( counts.destructor != 2 ) return Fail();
	if ( counts.object != 1 ) return Fail();
 }
	if ( counts.object != 0 ) return Fail();
	if ( counts.destructor != 3 ) return Fail();
 // ----------------------------


 // ----------------------------
 // Test SharedPtrs created as auto-vars.
 {
	CDEBUG << "Testing SharedPtr (auto var) with 3 references using copy constr...\n";
	counts.Clear();

	Object* obj = new Object;
	SharedPtr<Object> shared( obj );            // same/shared object
	SharedPtr<Object> shared2( shared );        // same/shared object
	SharedPtr<Object> shared3( shared );        // same/shared object

	if ( GetRefCnt( shared ) != 3 ) return Fail();		// 3 references to one shared object
	if ( GetRefCnt( shared2 ) != 3 ) return Fail();
	if ( GetRefCnt( shared3 ) != 3 ) return Fail();

	if ( shared  != obj ) return Fail();  // test SharedPtr::operator!=()
	if ( shared2 != obj ) return Fail();
	if ( shared3 != obj ) return Fail();
	if ( ! (shared  == obj) ) return Fail();  // test SharedPtr::operator==()
	if ( ! (shared2 == obj) ) return Fail();
	if ( ! (shared3 == obj) ) return Fail();

    // Test operator==/!= with identical SharedPtrs.
    if ( shared != shared2 ) return Fail();
    if ( ! (shared == shared2) ) return Fail();

	// Test accessing Object via SharedPtr.
	if ( shared->mTestVal != TEST_VAL ) return Fail();
	if ( shared->Magic() != MAGIC ) return Fail();
	if ( shared->mTestVal != TEST_VAL ) return Fail();
	if ( shared->Magic() != MAGIC ) return Fail();
	if ( (*shared).mTestVal != TEST_VAL ) return Fail();
	if ( (*shared).Magic() != MAGIC ) return Fail();

	if ( counts.object != 1 ) return Fail();
	if ( counts.constructor != 1 ) return Fail();
	if ( counts.destructor != 0 ) return Fail();
 }
    // SharedPtr auto-vars were deleted, check.
	if ( counts.object != 0 ) return Fail();		// counts should be the same
	if ( counts.destructor != 1 ) return Fail();
 // ----------------------------


 // ----------------------------
 // Test SharedPtrs created by new (pointer to SharedPtr).
 // (This form is useless outside testing: SharedPtrs themselves won't be pointers in practice)
 {
	CDEBUG << "Testing SharedPtr (ptr) with 2 references using copy constr...\n";
	counts.Clear();

	Object* obj = new Object;
	SharedPtr<Object>* shared  = new SharedPtr<Object>( obj );
	SharedPtr<Object>* shared2 = new SharedPtr<Object>( *shared );

	if ( GetRefCnt(shared)  != 2 ) return Fail();	// 2 references to one shared object
	if ( GetRefCnt(shared2) != 2 ) return Fail();

	if ( *shared  != obj ) return Fail();       // test SharedPtr::operator!=(T*obj)
	if ( *shared2 != obj ) return Fail();
	if ( ! (*shared  == obj) ) return Fail();   // test SharedPtr::operator==(T*obj)
	if ( ! (*shared2 == obj) ) return Fail();

	if ( counts.object != 1 ) return Fail();
	if ( counts.constructor != 1 ) return Fail();
	if ( counts.destructor != 0 ) return Fail();

	delete shared2;
	if ( GetRefCnt(shared) != 1 ) return Fail();	// reference count should be 1
	if ( counts.object != 1 ) return Fail();		// counts should be the same
	if ( counts.constructor != 1 ) return Fail();
	if ( counts.destructor != 0 ) return Fail();

	delete shared;
	if ( counts.object != 0 ) return Fail();		// counts should be the same
	if ( counts.destructor != 1 ) return Fail();
 }
 // ----------------------------


 // ----------------------------
 // Test SharedPtrs as class members.
 {
	CDEBUG << "Testing SharedPtr as class members...\n";
	counts.Clear();

    // See ctor.
    ClassWith3SharedPtrs c3;
    if ( c3.mSharedPtr1->mTestVal != 101 ) return Fail();
    if ( c3.mSharedPtr2->mTestVal != 102 ) return Fail();
    if ( c3.mSharedPtr3->mTestVal != 103 ) return Fail();

    // Test copying a class with SharedPtrs.
    Object* obj = new Object;
    ClassWith1SharedPtr c1a;
    c1a.mSharedPtr1 = obj;  // refCnt=1
    ClassWith1SharedPtr c1b;
    c1b = c1a;              // refCnt=2
    if ( c1a.mSharedPtr1.PTR() != obj ) return Fail();
    if ( c1b.mSharedPtr1.PTR() != obj ) return Fail();
    if ( GetRefCnt(c1a.mSharedPtr1) != GetRefCnt(c1b.mSharedPtr1) ) return Fail();
    if ( GetRefCnt(c1a.mSharedPtr1) != 2 ) return Fail();
    if ( GetRefCnt(c1b.mSharedPtr1) != 2 ) return Fail();
 }
 // ----------------------------


 // ----------------------------
 // Test SharedPtrs that point to different objects.
 {
	CDEBUG << "Testing SharedPtr to different objects...\n";
	counts.Clear();

	SharedPtr<Object> shared( new Object );
	SharedPtr<Object> shared2( new Object );

	if ( GetRefCnt(shared)  != 1 ) return Fail();  // 3 references to one shared object
	if ( GetRefCnt(shared2) != 1 ) return Fail();

    // Test operator==/!= with SharedPtrs that point to different actual objects
    if ( shared == shared2 ) return Fail();
    if ( ! (shared != shared2) ) return Fail();
 }
 // ----------------------------


 // ----------------------------
 // Test passing as SharedPtr to a function which copies it.
 {
   {
	CDEBUG << "Testing passing SharedPtr to a function...\n";
	counts.Clear();

	SharedPtr<Object> shared( new Object );  // refcnt = 1
	if ( GetRefCnt(shared) != 1 ) return Fail();
	if ( counts.object != 1 ) return Fail();
	if ( counts.constructor != 1 ) return Fail();
	if ( counts.destructor != 0 ) return Fail();

	// Pass SharedPtr to a function by value
	// which should increment the reference count.
	// Note SharedPtr isn't meant to be passed as a reference nor pointer.
	if ( not FuncPassedSharedPtr( shared, 1, 1 ) ) return Fail();
	if ( GetRefCnt(shared) != 1 ) return Fail();   // call func didn't keep its refcnt
	if ( counts.object != 1 ) return Fail();
	if ( counts.constructor != 1 ) return Fail();
	if ( counts.destructor != 0 ) return Fail();
   }
	if ( counts.object != 0 ) return Fail();
	if ( counts.constructor != 1 ) return Fail();
	if ( counts.destructor != 1 ) return Fail();
 }
 // ----------------------------
 {  // Function that is called keeps the SharedPtr.
   {
	counts.Clear();
	SharedPtr<Object> shared( new Object );  // refcnt = 1
	if ( not FuncPassedSharedPtrThatKeepsIt( shared ) ) return Fail();
	if ( GetRefCnt(shared) != 2 ) return Fail();
	if ( counts.object != 1 ) return Fail();
	if ( counts.constructor != 1 ) return Fail();
	if ( counts.destructor != 0 ) return Fail();
   }
    // Object should still exist even though the autovar SharedPtr was destroyed.
	if ( counts.object != 1 ) return Fail();
	if ( counts.constructor != 1 ) return Fail();
	if ( counts.destructor != 0 ) return Fail();
 }

 // ----------------------------
 // Test function that returns a SharedPtr
 {
   {
	CDEBUG << "Testing function that returns a SharedPtr...\n";
	counts.Clear();
	SharedPtr<Object> shared = FuncThatReturnsSharedPtr();
	if ( GetRefCnt(shared) != 1 ) return Fail();
	if ( counts.object != 1 ) return Fail();

	if ( shared.PTR() == NULL ) return Fail();
	if ( counts.object != 1 ) return Fail();  // PTR() should have no affect on ref-cnt
   }
	if ( counts.object != 0 ) return Fail();  // it should've been destroyed
 }

 {
   {
	counts.Clear();
	SharedPtr<Object> shared = FuncThatReturnsSharedPtr2();
	if ( GetRefCnt(shared) != 1 ) return Fail();
	if ( counts.object != 1 ) return Fail();
   }
	if ( counts.object != 0 ) return Fail();  // it should've been destroyed
 }

 // ----------------------------
 // Ensure that operator=(T*) is invoked rather than implict
 // conversion to SharedPtr<T> and invoking of operator=(SharedPtr<T)&)
 {
	CDEBUG << "Testing SharedPtr<T>::operator=(T*)...\n";
	counts.Clear();

	Object* obj = new Object;
	SharedPtr<Object> shared  = obj;
	SharedPtr<Object> shared2 = shared;
	SharedPtr<Object> shared3 = shared;
	SharedPtr<Object> shared4 = shared;

    // Preliminary checks before testing operator=(T*).
	if ( GetRefCnt(shared)  != 4 ) return Fail();	// 4 references to one shared object
	if ( GetRefCnt(shared2) != 4 ) return Fail();

    // This should invoke operator=(T*) which does mObj=obj.
    // A bug is if T* is implicitly converted to a SharedPtr<T>
    // and instead operator=(SharedPtr<T>/*fool checks.sh*/&) is applied!

    Object* obj2 = new Object;
    shared2 = obj2;

    if ( shared2.mObj != obj2 ) return Fail();

    // shared2 should have a reference count of 1,
    // because it is the only SharedPtr pointing to obj2.

    if ( GetRefCnt(shared2) != 1 ) return Fail();

    // The other 3 SharedPtr should be affected,
    // and have ref cnt = 3.

    if ( GetRefCnt(shared)  != 3 ) return Fail();
    if ( GetRefCnt(shared3) != 3 ) return Fail();
    if ( GetRefCnt(shared4) != 3 ) return Fail();

    // Ref cnt pointer of shared2 should differ.
    if ( GetRefCnt(shared2) == GetRefCnt(shared) ) return Fail();
    if ( GetRefCnt(shared2) == GetRefCnt(shared3) ) return Fail();
    if ( GetRefCnt(shared2) == GetRefCnt(shared4) ) return Fail();
 }
 // ----------------------------


 // ----------------------------
 // Yet more tests of operator=(SharedPtr<T)&)
 {
	CDEBUG << "Testing SharedPtr<T>::operator=(SharedPtr)...\n";
	counts.Clear();

	Object* obj = new Object;
	SharedPtr<Object> shared  = obj;
	SharedPtr<Object> shared2 = shared;
	SharedPtr<Object> shared3 = shared2;

	if ( GetRefCnt(shared)  != 3 ) return Fail();	// 3 references to one shared object
	if ( GetRefCnt(shared2) != 3 ) return Fail();
    if ( GetRefCnt(shared3) != 3 ) return Fail();
 }
 // ----------------------------


 // ----------------------------
 // Test using SharedPtr with STL: vector<SharedPtr>.resize().
 // STL, depending on implementation, can call the SharedPtr's default and even copy constructor.
 {
   {
	CDEBUG << "Testing STL vector<SharedPtr>.resize()...\n";
	counts.Clear();

	vector<SharedPtr<Object> > vec;
	const int cnt = 50;
	vec.resize( cnt );		// resize() is what really allocates elements
	vec.reserve( cnt );		// test this along the way
	if ( vec.size() != (uint)cnt ) return Fail();

	Object* obj = new Object;
	SharedPtr<Object> shared( obj );

    for ( int i = 0; i < cnt; ++i )
	{
		vec[i] = shared;

		if ( vec[i] != shared ) return Fail();
		if ( vec[i].CONST_PTR() != shared.CONST_PTR() ) return Fail();

		if ( GetRefCnt(shared) != i+2 ) return Fail();	// when i=0, ref cnt will be 2
		if ( GetRefCnt(vec[i]) != i+2 ) return Fail();

		if ( vec[i]->Magic() != MAGIC ) return Fail();
	}
	if ( GetRefCnt(shared) != cnt+1 ) return Fail();		// +1 for original "shared"

   }
}
 // ----------------------------


 // ----------------------------
 // Test using STL vector and push_back().
 {
	const int cnt = 50;
   {
	CDEBUG << "Testing STL vector with push_back()...\n";
	counts.Clear();

	vector<SharedPtr<Object> > vec;

	Object* obj = new Object;
	SharedPtr<Object> shared( obj );

	int expectedRefCnt = 1;

    for ( int i = 0; i < cnt; ++i )
	{
		int prevRefCnt = GetRefCnt(shared);

		vec.push_back( shared );
		++expectedRefCnt;

		if ( vec[vec.size()-1] != shared ) return Fail();
		if ( vec[vec.size()-1]->Magic() != MAGIC ) return Fail();

		// After push_back(), ref cnt should've increased.
		if ( GetRefCnt(shared) <= prevRefCnt ) return Fail();  // ref cnt should've inc'd
		if ( GetRefCnt(vec[vec.size()-1]) != GetRefCnt(shared) ) return Fail();

		// Ref cnt should increase by one every time.
		if ( GetRefCnt(vec[vec.size()-1]) != expectedRefCnt ) return Fail();
		if ( GetRefCnt(shared)            != expectedRefCnt ) return Fail();

		if ( counts.object != 1 ) return Fail();  // there's one object
	}
   }
	if ( counts.object != 0 ) return Fail();  // it should've been deleted
	if ( counts.constructor != 1 ) return Fail();  // should've been only one
	if ( counts.destructor != 1 ) return Fail();
}
 // ----------------------------


 // ----------------------------
 // Test using STL vector and push_back() (another case).
 {
	const int cnt = 50;
   {
	CDEBUG << "Testing STL vector with push_back( SharedPtr<Object>( new Object ) )...\n";
	counts.Clear();

	vector<SharedPtr<Object> > vec;

    for ( int i = 0; i < cnt; ++i )
	{
		int prevObjectCnt = counts.object;

		vec.push_back( SharedPtr<Object>( new Object ) );

		if ( counts.object != prevObjectCnt+1 ) return Fail();
		if ( GetRefCnt(vec[vec.size()-1]) != 1 ) return Fail();  // pushed value makes 1 ref
		if ( vec[vec.size()-1]->Magic() != MAGIC ) return Fail();
	}
   }
	if ( counts.object != 0 ) return Fail();  // all should've been deleted
	if ( counts.constructor != cnt ) return Fail();  // should've been cnt instances
	if ( counts.destructor != cnt ) return Fail();
}
 // ----------------------------


 // ----------------------------
 // Test using a huge STL vector.
 {
	CDEBUG << "Testing huge STL vector<SharedPtr> to induce reallocation...\n";
	counts.Clear();
	const int cnt = 500*1000;  // induce STL vector to reallocate
	SharedPtr<Object> sharedObj = new Object;
  {
	vector<SharedPtr<Object> > vec;
	vec.reserve( 10 );  // induce STL vector to reallocate
    void* first = NULL;
    for ( int i = 0; i < cnt; ++i )
	{
		vec.push_back( sharedObj  );
		if ( i == 0 ) first = (void*) &vec[0];
	}
	// There's one shared object and it has cnt+1 reference counts.
	if ( GetRefCnt(sharedObj) != cnt + 1 ) return Fail();
	if ( counts.constructor != 1 ) return Fail();
	if ( counts.object      != 1 ) return Fail();

	// STL should have reallocated the vector, else this test case was ineffective.
	if ( (void*)&vec[0] == first ) return Fail();
  }
	// STL vector was destroyed.  One refcnt remains at "sharedObj".
	if ( GetRefCnt(sharedObj) != 1 ) return Fail();
	if ( counts.object != 1 ) return Fail();
 }
    // Now last shared object was destroyed.
	if ( counts.constructor != 1 ) return Fail();
	if ( counts.destructor != 1 ) return Fail();
	if ( counts.object != 0 ) return Fail();

 // ----------------------------


 // ----------------------------
 // Test finding/removing a SharedPtr in a STL list.
 {
	CDEBUG << "Testing STL list<SharedPtr>...\n";

	Int::Clear();
	list<SharedPtr<Int> > nums;

	SharedPtr<Int> one   = new Int(1);
	SharedPtr<Int> two   = new Int(2);
	SharedPtr<Int> three = new Int(3);
	if ( Int::msInstanceCnt != 3 ) { Fail(); }
	if ( one.GetRefCnt() != 1 ) { Fail(); }
	if ( two.GetRefCnt() != 1 ) { Fail(); }
	if ( three.GetRefCnt() != 1 ) { Fail(); }

	nums.push_front( one );
	nums.push_front( two );
	nums.push_front( three );
	if ( Int::msInstanceCnt != 3 ) { Fail(); }
	if ( one.GetRefCnt() != 2 ) { Fail(); }
	if ( two.GetRefCnt() != 2 ) { Fail(); }
	if ( three.GetRefCnt() != 2 ) { Fail(); }

	// Erase two.
	list<SharedPtr<Int> >::iterator iter = std::find( nums.begin(), nums.end(), two );
	if ( iter == nums.end() ) { return Fail(); }
	if ( (*iter)->Get() != 2 ) { return Fail(); }
	nums.erase( iter );
	if ( one.GetRefCnt() != 2 ) { Fail(); }
	if ( two.GetRefCnt() != 1 ) { Fail(); }  // erased two, minus one ref
	if ( three.GetRefCnt() != 2 ) { Fail(); }
 }
	if ( Int::msInstanceCnt != 0 ) { Fail(); }
 // ----------------------------


 // ----------------------------
 // Test self-assignment.
 {
	CDEBUG << "Testing self-assignment...\n";
	counts.Clear();
  {
	SharedPtr<Object> obj( new Object );
	obj = obj;
	if ( GetRefCnt(obj) != 1 ) return Fail();
  }

  // Test operator=(Shared*) ability to handle a form of self-assignment.
  {
	Object* obj = new Object;
	SharedPtr<Object> sharedPtr;
	sharedPtr = obj;
	sharedPtr = obj; if ( sharedPtr.mObj == NULL ) return Fail();  // test if deleted
	if ( GetRefCnt(*obj) != 1 ) return Fail();
	sharedPtr = obj;  // self-assignment doesn't effectively change ref-cnt,
	sharedPtr = obj;  // though SharedPtr::operator=() increments and decrements it
	sharedPtr = obj;
	if ( GetRefCnt(*obj) != 1 ) return Fail();
  }
 }
 // ----------------------------

 // ----------------------------
 // Test converting a raw pointer to Shared back to a SharedPtr.
 {
	CDEBUG << "Testing converting back to a SharedPtr...\n";
	counts.Clear();
  {
	SharedPtr<Object> obj( new Object );
	obj = obj;

    Object* raw = obj.PTR();
    SharedPtr<Object> obj2 = raw;

	if ( GetRefCnt(obj) != 2 ) return Fail();
  }
 }
 // ----------------------------

 // ============
	return true;
}

/*****************************************************************************
 * Test SharedPtrNull.
 *****************************************************************************/
class Dummy : public Shared
{
};

bool
TestSharedPtrNull( void )
{
	CDEBUG << "Testing SharedPtrNull...\n";

	SharedPtrNull<Dummy> dummyDefaultCtor;  // SharedPtr(void)
	SharedPtrNull<Dummy> dummy = NULL;      // SharedPtr(SHARED*)
	SharedPtrNull<Dummy> dummy2 = dummy;    // SharedPtr copy ctor

	if ( dummy != NULL ) { Fail(); }
	if ( dummy2 != NULL ) { Fail(); }

	dummy = new Dummy;
	if ( dummy == NULL ) { Fail(); }
	if ( dummy.GetRefCnt() != 1 ) { Fail(); }

	dummy = NULL;
	if ( dummy != NULL ) { Fail(); }

  {
	Dummy* pDummy1 = new Dummy;
	Dummy* pDummy2 = new Dummy;
	if ( pDummy1->GetRefCnt() || pDummy2->GetRefCnt() ) { Fail(); }
	SharedPtrNull<Dummy> shptrDummy1;
	SharedPtrNull<Dummy> shptrDummy2;
	shptrDummy1 = pDummy2;
	shptrDummy2 = pDummy1;
	if ( pDummy1->GetRefCnt() != 1 || pDummy2->GetRefCnt() != 1 ) { Fail(); }
  }

	return true;
}

/*****************************************************************************
 * [2007/05 ]
 * This caught a deadlock bug in ~SharedPtr().
 * ~SharedPtr() held a lock, called "delete SharedPtr" which called ~SharedPtr() -- deadlock!
 *
 * (gdb) bt
 * #0  0x0000000000406ce1 in base::Atomic::Xchg (ptr=0x60d9c0, val=1)
 *     at ../../base/src/base_thread_atomic.hh:114
 * #1  0x0000000000406e24 in base::RushLock::Lock (this=0x60d9c0)
 *     at ../../base/src/base_thread_lock.hh:181
 * #2  0x0000000000406f56 in ~SharedPtr (this=0x60ef40)
 *     at ../../base/src/base_shared_ptr2.hh:83
 * #3  0x0000000000407068 in ~Node (this=0x60ef30) at src/Atomic.cc:90
 * #4  0x00000000004070f7 in ~Node2 (this=0x60ef30) at src/Atomic.cc:98
 * #5  0x0000000000407228 in ~SharedPtr (this=0x7fff7effe750)
 *     at ../../base/src/base_shared_ptr2.hh:86
 * #6  0x0000000000406825 in Test2 () at src/Atomic.cc:107
 * #7  0x0000000000406bfa in main (argc=1, argv=0x7fff7effe8a8)
 *     at src/Atomic.cc:117
 *
 * Buggy code:
 * ~SharedPtr()
 * {
 * SHARED_PTR_LOCK();
 * CHECK(*this);
 *     if ( --(mObj->mRefCnt) == 0 )  // delete object if last reference
 *          delete mObj;
 * SHARED_PTR_UNLOCK();
 * }
 *
 * Corrected (BTW, NULL check added for unrelated reason):
 * ~SharedPtr()
 * {
 *     // Necessary evil in case SharedPtr(INVALID) stayed invalid.
 *                               ^^^^^^^^^^^^^^^^^^ replaced by SharedPtrNull
 *     if ( UX( mObj == NULL ) )
 *          return;
 *
 * SHARED_PTR_LOCK();
 *     CHECK(*this);
 *     int refCnt = --(mObj->mRefCnt);
 * SHARED_PTR_UNLOCK(); // unlock since delete may invoke another ~SharedPtr()
 *     if ( refCnt == 0 )  // delete object if last reference
 *          delete mObj;
 * }
 *****************************************************************************/
class Data : public Shared
{
// nothing
};

class Node : public Shared
{
public:
	Node( void ) : mData(NULL), mData2(new Data) { }
	SharedPtrNullLocked<Data>	mData;
	SharedPtrNullLocked<Data>	mData2;
};

class Node2 : public Node
{
public:
// nothing
};

bool
TestSharedPtrNull2( void )
{
	SharedPtrNullLocked<Node2> node2 = new Node2;
	return true;
}

/*****************************************************************************
 * Test deadlock case (was encountered as a bug).
 *
 * This deadlock happened because a destructor has the global SharedPtr lock
 * and nested destructor calls SharedPtr::operator=() which deadlocks.
 *****************************************************************************/
class Deadlock0 : public Shared
{
};

class Deadlock1 : public Shared
{
public:
	Deadlock1( void )
	{
		mPtr = new Deadlock0;
	}

public:
	SharedPtrNullLocked<Deadlock0>  mPtr;
};

class Deadlock2
{
public:
	Deadlock2( void )
	{
		mPtr = new Deadlock1;
	}

	~Deadlock2()
	{
		// SharedPtr::operator=() in a destructor like this once was a bug.
		mPtr = NULL;
	}

public:
	SharedPtrNullLocked<Deadlock1>  mPtr;
};

bool
TestDeadlock( void )
{
	CDEBUG << "Testing deadlock case...\n";
	Deadlock2 deadlock2;
	return true;  // or test will deadlock
}

/*****************************************************************************
 * Test that a Shared object can be interchanged with different kinds of SharedPtr.
 *****************************************************************************/
bool
TestInterchangingSharedPtr( void )
{
 {
	CDEBUG << "Testing interchanging Shared...\n";
	counts.Clear();

	SharedPtr<Object> obj = new Object;
	if ( counts.object != 1 ) return Fail();
	if ( counts.constructor != 1 ) return Fail();
	if ( obj.GetRefCnt() != 1 ) return Fail();

	SharedPtrNull<Object> obj_ = obj.PTR();
	if ( counts.object != 1 ) return Fail();
	if ( counts.constructor != 1 ) return Fail();
	if ( obj.GetRefCnt() != 2 ) return Fail();
 }
	return true;
}

/*****************************************************************************
 * Test copying Shared.
 *****************************************************************************/
class DerivedShared : public Shared
{
};

bool
TestCopyingShared( void )
{
	CDEBUG << "Testing copying Shared...\n";
	shptr<DerivedShared> obj1 = new DerivedShared;  // refcnt=1
	DerivedShared        obj2;                      // refcnt=0

	// Will obj1's refcnt be overwritten with 0?  It should be 1.
	*obj1 = obj2;  // copy Shared = Shared
	if ( obj1->GetRefCnt() != 1 ) { Fail(); }
	if ( obj2.GetRefCnt() != 0 ) { Fail(); }
	return true;
}

/*****************************************************************************
 * Test SharedPtr polymorphism: ability to pass Shared<Derived> as Shared<Base>.
 *****************************************************************************/
class Base_Shared : public Shared  // running out of names
{
public:
	CLASS_CONST uint BASE_SIG = 0x1234aaaa;
	Base_Shared( void ) : mBaseSig(BASE_SIG) { }
	uint mBaseSig;
};

class Derived_Shared : public Base_Shared
{
public:
	CLASS_CONST uint DERIVED_SIG = 0x99887766;
	Derived_Shared( void ) : mDerivedSig(DERIVED_SIG) { }
	uint mDerivedSig;
};

SharedPtr<Base_Shared>
ReflectBaseShared( SharedPtr<Base_Shared> base )
{
	if ( base->mBaseSig != Base_Shared::BASE_SIG ) { Fail(); }
	if ( base->GetRefCnt() != 2 ) { Fail(); return NULL; }
	return base;
}

bool
TestBaseDerivedSharedPtr( void )
{
 {
	CDEBUG << "Testing passing Shared<Derived> as Shared<Base>...\n";
	shptr<Derived_Shared> derived = new Derived_Shared;  // refcnt=1

	if ( derived->GetRefCnt() != 1 ) { return Fail(); }
	shptr<Base_Shared> base = ReflectBaseShared( derived );
	if ( derived->GetRefCnt() != 2 ) { return Fail(); }  // assignment to base incs it
	if (    base->GetRefCnt() != 2 ) { return Fail(); }  // base=derived (same object)
	Derived_Shared* derivedRaw = static_cast<Derived_Shared*>( base.PTR() );
	if ( derivedRaw->mDerivedSig != Derived_Shared::DERIVED_SIG ) { return Fail(); }
 }

#if 0 // This is of course illegal and should not, and does not, compile (gcc 4.1).
 {
	shptr<Base_Shared>    base( new Base_Shared );
	shptr<Derived_Shared> derived( base );
 }
#endif

	return true;
}

/*****************************************************************************
 * main.
 *****************************************************************************/
int
main( int argc, char** argv )
{
	TEST_MAIN_BEGIN;
	int res = 1;
	if ( TestSharedPtr()
      && TestSharedPtrNull()
      && TestSharedPtrNull2()
      && TestDeadlock()
      && TestInterchangingSharedPtr()
      && TestCopyingShared()
      && TestBaseDerivedSharedPtr()
       )
	{
		CDEBUG << "Test PASSED. \n";
		res = 0;
	}
	else
	{
		CDEBUG << "Test FAILED! \n";
		res = 1;
	}
	TEST_MAIN_END;
	exit( res );
}
